# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module handles random number generation.

It derives most of the random number procedures from `RandomState` in
`numpy.random` and defines a class which used `RandomState` to generate
pseudo-random numbers.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

RandomGeneratorBase (:py:class:`.RandomGeneratorBase`)
    An abstract base class for random number generators.

RandomGenerator (:py:class:`.RandomGenerator`)
    A class representing a random number generator.

ReservoirSampler (:py:class:`.ReservoirSampler`)
    A class for reservoir sampling.

MockRandomGenerator (:py:class:`.MockRandomGenerator`)
    A non-random number generator which is useful for testing.
"""
from abc import ABCMeta, abstractmethod
import logging
import numpy as np
from numpy.random import RandomState
from pyretis.core.particlefunctions import (calculate_kinetic_temperature,
                                            reset_momentum)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = [
    'RandomGeneratorBase',
    'RandomGenerator',
    'ReservoirSampler',
    'MockRandomGenerator',
    'create_random_generator'
]


class RandomGeneratorBase(metaclass=ABCMeta):
    """A base class for random number generators.

    This is a base class for random number generators. It does not
    actually implement a generator.

    Attributes
    ----------
    seed : int
        A seed for the generator

    """

    def __init__(self, seed=0):
        """Initialise the random number generator.

        Parameters
        ----------
        seed : int, optional
            An integer used for seeding the generator if needed.

        """
        self.seed = seed

    @abstractmethod
    def rand(self, shape=1):
        """Draw random numbers in [0, 1).

        Parameters
        ----------
        shape : int, optional
            The number of numbers to draw

        Returns
        -------
        out : float
            Pseudo-random number in [0, 1)

        """
        return

    @abstractmethod
    def get_state(self):
        """Return info about the current state."""
        return

    @abstractmethod
    def set_state(self, state):
        """Set state for random generator."""
        return

    @abstractmethod
    def random_integers(self, low, high):
        """Draw random integers in [low, high].

        Parameters
        ----------
        low : int
            This is the lower limit
        high : int
            This is the upper limit

        Returns
        -------
        out : int
            The pseudo-random integers in [low, high].

        """
        return

    @abstractmethod
    def normal(self, loc=0.0, scale=1.0, size=None):
        """Return values from a normal distribution.

        Parameters
        ----------
        loc : float, optional
            The mean of the distribution
        scale : float, optional
            The standard deviation of the distribution
        size : int, tuple of ints, optional
            Output shape, i.e. how many values to generate. Default is
            None which is just a single value.

        Returns
        -------
        out : float, numpy.array of floats
            The random numbers generated.

        """
        return

    @abstractmethod
    def multivariate_normal(self, mean, cov, cho=None, size=1):
        """Draw numbers from a multi-variate distribution.

        Parameters
        ----------
        mean : numpy array (1D, 2)
            Mean of the N-dimensional array
        cov : numpy array (2D, (2, 2))
            Covariance matrix of the distribution.
        cho : numpy.array (2D, (2, 2)), optional
            Cholesky factorization of the covariance. If not given,
            it will be calculated here.
        size : int, optional
            The number of samples to do.

        Returns
        -------
        out : float or numpy.array of floats size
            The random numbers that are drawn here.

        """
        return

    def generate_maxwellian_velocities(self, particles, boltzmann, temperature,
                                       dof, selection=None, momentum=True):
        """Generate velocities from a Maxwell distribution.

        The velocities are drawn to match a given temperature and this
        function can be applied to a subset of the particles.

        The generation is done in three steps:

        1) We generate velocities from a standard normal distribution.

        2) We scale the velocity of particle `i` with
           ``1.0/sqrt(mass_i)`` and reset the momentum.

        3) We scale the velocities to the set temperature.

        Parameters
        ----------
        particles : object like :py:class:`.Particles`
            These are the particles to set the velocity of.
        boltzmann : float
            The Boltzmann factor in correct units.
        temperature : float
            The desired temperature.
            Typically, `system.temperature['set']` will be used here.
        dof : list of floats, optional
            The degrees of freedom to subtract. Its shape should
            be equal to the number of dimensions.
        selection : list of ints, optional
            A list with indices of the particles to consider.
            Can be used to only apply it to a selection of particles
        momentum : boolean, optional
            If true, we will reset the momentum.

        Returns
        -------
        out : None
            Returns `None` but modifies velocities of the selected
            particles.

        """
        if selection is None:
            vel, imass = particles.vel, particles.imass
        else:
            vel, imass = particles.vel[selection], particles.imass[selection]
        vel = np.sqrt(imass) * self.normal(loc=0.0, scale=1.0,
                                           size=vel.shape)
        # NOTE: x[None] = x for a numpy.array - this is not valid for a list.
        particles.vel[selection] = vel
        if momentum:
            reset_momentum(particles, selection=selection)

        _, avgtemp, _ = calculate_kinetic_temperature(particles, boltzmann,
                                                      dof=dof,
                                                      selection=selection)
        scale_factor = np.sqrt(temperature/avgtemp)
        particles.vel[selection] *= scale_factor

    def draw_maxwellian_velocities(self, system, sigma_v=None):
        """Draw numbers from a Gaussian distribution.

        Parameters
        ----------
        system : object like :py:class:`.System`
            This is used to determine the temperature parameter(s) and
            the shape (number of particles and dimensionality)
        sigma_v : numpy.array, optional
            The standard deviation in velocity, one for each particle.
            If it's not given it will be estimated.

        """
        if not sigma_v or sigma_v < 0.0:
            kbt = 1.0/system.temperature['beta']
            # sigma_v is (n, 1) matrix
            sigma_v = np.sqrt(kbt*system.particles.imass)
        npart, dim = system.particles.vel.shape
        vel = self.normal(loc=0.0, scale=sigma_v, size=(npart, dim))
        return vel, sigma_v


class RandomGenerator(RandomGeneratorBase):
    """A random number generator from numpy.

    This class that defines a random number generator. It will use
    `numpy.random.RandomState` for the actual generation and we refer
    to the numpy documentation [1]_.

    Attributes
    ----------
    seed : int
        A seed for the pseudo-random generator.
    rgen : object like :py:class:`numpy.random.RandomState`
        This is a container for the Mersenne Twister pseudo-random
        number generator as implemented in numpy [#]_.

    References
    ----------
    .. [#] The NumPy documentation on RandomState,
       http://docs.scipy.org/doc/numpy/reference/generated/numpy.random.RandomState.html

    """

    def __init__(self, seed=0):
        """Initialise the random number generator.

        If a seed is given, the random number generator will be seeded.

        Parameters
        ----------
        seed : int, optional
            An integer used for seeding the generator if needed.

        """
        super().__init__(seed=seed)
        self.rgen = RandomState(seed=seed)

    def choice(self, a, size=None, replace=True, p=None):
        """Chooses random samples.

        Parameters
        ----------
        a : iterable, int
            The group to choose from, an int is converted to range(a).
        size : int, optional
            The number of samples to choose.
        replace : bool, optional
            If to replace samples between draws, default True.
        p : iterable, optional
            The probabilities for every option in a,
            default is an uniform dirstribution.

        Returns
        -------
        choice : array-like
            The picked choices.

        """
        return self.rgen.choice(a, size, replace, p)

    def rand(self, shape=1):
        """Draw random numbers in [0, 1).

        Parameters
        ----------
        shape : int, optional
            The number of numbers to draw

        Returns
        -------
        out : float
            Pseudo-random number in [0, 1)

        Note
        ----
        Here, we will just draw a list of numbers and not for
        an arbitrary shape.

        """
        return self.rgen.rand(shape)

    def get_state(self):
        """Return current state."""
        state = {'seed': self.seed,
                 'state': self.rgen.get_state(),
                 'rgen': 'rgen'}
        return state

    def set_state(self, state):
        """Set state for random generator."""
        return self.rgen.set_state(state['state'])

    def random_integers(self, low, high):
        """Draw random integers in [low, high].

        Parameters
        ----------
        low : int
            This is the lower limit
        high : int
            This is the upper limit

        Returns
        -------
        out : int
            The pseudo-random integers in [low, high].

        Note
        ----
        np.random.randint(low, high) is defined as drawing
        from `low` (inclusive) to `high` (exclusive).

        """
        return self.rgen.randint(low, high + 1)

    def normal(self, loc=0.0, scale=1.0, size=None):
        """Return values from a normal distribution.

        Parameters
        ----------
        loc : float, optional
            The mean of the distribution
        scale : float, optional
            The standard deviation of the distribution
        size : int, tuple of ints, optional
            Output shape, i.e. how many values to generate. Default is
            None which is just a single value.

        Returns
        -------
        out : float, numpy.array of floats
            The random numbers generated.

        """
        return self.rgen.normal(loc=loc, scale=scale, size=size)

    def multivariate_normal(self, mean, cov, cho=None, size=1):
        """Draw numbers from a multi-variate distribution.

        This is an attempt on speeding up the call of
        `RandomState.multivariate_normal` if we need to call it over and
        over again. Such repeated calling will do an SVD repeatedly,
        which is wasteful. In this function, this transform can be
        supplied and it is only estimated if it's not explicitly given.

        Parameters
        ----------
        mean : numpy array (1D, 2)
            Mean of the N-dimensional array
        cov : numpy array (2D, (2, 2))
            Covariance matrix of the distribution.
        cho : numpy.array (2D, (2, 2)), optional
            Cholesky factorization of the covariance. If not given,
            it will be calculated here.
        size : int, optional
            The number of samples to do.

        Returns
        -------
        out : float or numpy.array of floats size
            The random numbers that are drawn here.

        See also
        --------
        numpy.random.multivariate_normal

        """
        if cho is None:
            cho = np.linalg.cholesky(cov)
        norm = self.normal(loc=0.0, scale=1.0, size=2*size)
        norm = norm.reshape(size, 2)
        meanm = np.array([mean, ] * size)
        return meanm + np.dot(norm, cho.T)


class ReservoirSampler:
    """A class representing a reservoir sampler.

    The reservoir sampler will maintain a list of `k` items drawn
    randomly from a set of `N > k` items. The list is created and
    maintained so that we only need to store `k`items This is useful
    when `N` is very large or when storing all `N` items require a lot
    of memory. The algorithm is described by Knuth [#]_ but here we do
    a variation, so that each item may be picked several times.


    Attributes
    ----------
    rgen : object like :py:class:`numpy.random.RandomState`
        This is a container for the Mersenne Twister pseudo-random
        number generator as implemented in numpy, see the documentation
        of `RandomGenerator`.
    items : integer
        The number of items seen so far, i.e. the current `N`.
    reservoir : list
        The items we have stored.
    length : integer
        The maximum number of items to store in the reservoir
        (i.e. `k`).
    ret_idx : integer
        This is the index of the item to return if we are requesting
        items from the reservoir.

    References
    ----------
    .. [#] The Art of Computer Programming.

    """

    def __init__(self, seed=0, length=10, rgen=None):
        """Initialise the reservoir.

        Parameters
        ----------
        seed : int, optional
            An integer used for seeding the generator.
        length : int, optional
            The maximum number of items to store.
        rgen : object like :py:class:`.RandomGenerator`, optional
            In case we want to re-use a random generator object.
            If this is specified, the parameter `seed` is ignored.

        """
        if rgen is not None:
            self.rgen = rgen
        else:
            self.rgen = RandomState(seed=seed)
        self.items = 0
        self.reservoir = []
        self.length = length
        self.ret_idx = 0

    def append(self, new_item):
        """Try to add an item to the reservoir.

        Parameters
        ----------
        new_item : any type
            This is the item we try to add to the reservoir.

        """
        self.items += 1
        if self.items == 1:
            self.reservoir = [new_item for _ in range(self.length)]
        else:
            factor = 1.0/float(self.items)
            for i in range(self.length):
                if self.rgen.rand() < factor:
                    self.reservoir[i] = new_item

    def get_item(self):
        """Return the next item from the reservoir.

        Returns
        -------
        out : any type
            Returns an item from the reservoir.

        """
        if self.ret_idx >= self.length:
            self.ret_idx = 0
            msg = ['Out of bounds in the reservoir sampler!']
            msg += ['Please increase the size of the reservoir.']
            msgtxt = '\n'.join(msg)
            logger.critical(msgtxt)
        ret = self.reservoir[self.ret_idx]
        self.ret_idx += 1
        return ret


class MockRandomGenerator(RandomGeneratorBase):
    """A **mock** random generator, useful **only for testing**.

    This class represents a random generator that can be used for
    testing algorithms. It will simply return numbers from a
    small list of pseudo-random numbers. This class is only useful
    for testing algorithms on different systems. It should *NEVER*
    be used for actual production runs!
    """

    def __init__(self, seed=0, norm_shift=False):
        """Initialise the mock random number generator.

        Here, we set up some predefined random number which we will
        use as a pool for the generation.

        Parameters
        ----------
        seed : int, optional
            An integer used for seeding the generator if needed.
        norm_shift: boolean
            If True is will ensure that the fake 'normal distribution'
            is shifted to get the right mean.

        """
        super().__init__(seed=seed)
        self.rgen = [0.78008018, 0.04459916, 0.76596775, 0.97676713,
                     0.53799598, 0.98657116, 0.36343553, 0.55356511,
                     0.03172585, 0.48984682, 0.73416687, 0.98453452,
                     0.55129902, 0.40598753, 0.59448394, 0.26823255,
                     0.31168372, 0.05072849, 0.44876368, 0.94301709]
        self.length = len(self.rgen)
        self.randint = self.seed
        self.norm_shift = norm_shift
        logger.critical('You are using a "mock" random generator!\n')
        if norm_shift:
            logger.critical('Fake-normal is shifted.\n')
            logger.critical('Comparison with TISMOL might fail\n')
        else:
            logger.critical('Fake-normal is not shifted.\n')
            logger.critical('Random numbers not centered around 0.\n')

    def rand(self, shape=1):
        """Draw random numbers in [0, 1).

        Parameters
        ----------
        shape : int
            The number of numbers to draw.

        Returns
        -------
        out : float
            Pseudo-random number in [0, 1).

        """
        numbers = []
        for _ in range(shape):
            if self.seed >= self.length:
                self.seed = 0
            numbers.append(self.rgen[self.seed])
            self.seed += 1
        return np.array(numbers)

    def get_state(self):
        """Return current state."""
        state = {'seed': self.seed,
                 'state': self.seed,
                 'rgen': 'mock'}
        return state

    def set_state(self, state):
        """Set current state."""
        self.seed = state['state']

    def random_integers(self, low, high):
        """Return random integers in [low, high].

        Parameters
        ----------
        low : int
            This is the lower limit
        high : int
            This is the upper limit

        Returns
        -------
        out : int
            This is a pseudo-random integer in [low, high].

        """
        idx = self.rand()[0]*(high-low+1)
        return int(idx) + low

    def normal(self, loc=0.0, scale=1.0, size=None):
        """Return values from a normal distribution.

        Parameters
        ----------
        loc : float, optional
            The mean of the distribution
        scale : float, optional
            The standard deviation of the distribution
        size : int, tuple of ints, optional
            Output shape, i.e. how many values to generate. Default is
            None which is just a single value.

        Returns
        -------
        out : float, numpy.array of floats
            The random numbers generated.

        Note
        ----
        This is part of the Mock-random number generator. Hence, it
        won't provide a true normal distribution, though the mean is set to
        the value of loc.

        """
        if self.norm_shift:
            shift = loc-0.5
        else:
            shift = 0.
        if size is None:
            return self.rand(shape=1)+shift
        numbers = np.zeros(size)
        for i in np.nditer(numbers, op_flags=['readwrite']):
            i[...] = self.rand(shape=1)[0] + shift
        return numbers

    def multivariate_normal(self, mean, cov, cho=None, size=1):
        """Draw numbers from a multi-variate distribution.

        This is an attempt on speeding up the call of
        `RandomState.multivariate_normal` if we need to call it over and
        over again. Such repeated calling will do an SVD repeatedly,
        which is wasteful. In this function, this transform can be
        supplied and it is only estimated if it's not explicitly given.

        Parameters
        ----------
        mean : numpy array (1D, 2)
            Mean of the N-dimensional array
        cov : numpy array (2D, (2, 2))
            Covariance matrix of the distribution.
        cho : numpy.array (2D, (2, 2)), optional
            Cholesky factorization of the covariance. If not given,
            it will be calculated here.
        size : int, optional
            The number of samples to do.

        Returns
        -------
        out : float or numpy.array of floats size
            The random numbers that are drawn here.

        See also
        --------
        numpy.random.multivariate_normal

        """
        norm = self.normal(loc=0.0, scale=1.0, size=2*size)
        norm = norm.reshape(size, 2)
        meanm = np.array([mean, ] * size)
        return 0.01*(meanm + norm)

    def choice(self, a, size=None, replace=True, p=None):
        """Choose random samples.

        Parameters
        ----------
        a : iterable
            The group to choose from.
        size : int, optional
            The number of samples to choose.
        replace : bool, optional
            If to replace samples between draws, default True.
        p : iterable, optional
            The probabilities for every option in a,
            default is an uniform dirstribution.

        Returns
        -------
        choice : array-like
            The picked choices.

        """
        if isinstance(a, int):
            a = list(range(a))
        if p is None:
            p = [1/len(a) for i in a]
        if size is None:
            size = 1
        out = []
        for _ in range(size):
            rand = self.rand()
            prob0 = 0
            # Make sure p sums to 1 even after popping
            p_sum = sum(p)
            p = [i/p_sum for i in p]
            for i, probi in zip(a, p):
                prob0 += probi
                if rand < prob0:
                    out.append(i)
                    if not replace:
                        idx = a.index(i)
                        _ = a.pop(idx)
                        _ = p.pop(idx)
                    break
        return out


class Borg:
    """A class for sharing states of objects."""

    class_state = None
    number_of_borgs = 0

    @classmethod
    def update_state(cls, state):
        """Update the class state and enable sharing of it."""
        if cls.class_state is None:
            logger.debug('Setting state to the shared.')
            cls.class_state = state
        else:
            logger.debug('Reusing shared state.')
        return cls.class_state

    @classmethod
    def reset_state(cls):
        """Remove memory of the shared state."""
        cls.class_state = None

    @classmethod
    def make_new_swarm(cls):
        """Make new swarm (new class)."""
        name = cls.__name__ + str(cls.number_of_borgs)
        cls.number_of_borgs += 1
        new_borg = type(name, cls.__bases__, dict(cls.__dict__))
        new_borg.reset_state()  # pylint: disable=no-member
        return new_borg

    def __init__(self, *args, **kwargs):
        """Enable share of the state.

        This will update the Borg.class_state and
        update self.__dict__ to this object, which enables
        sharing of the state.

        """
        super().__init__(*args, **kwargs)
        self.__dict__ = self.update_state(self.__dict__)


class RandomGeneratorBorg(Borg, RandomGenerator):
    """A class for sharing the state between RandomGenerator objects."""


class MockRandomGeneratorBorg(Borg, MockRandomGenerator):
    """A class for sharing the state between MockRandomGenerator objects."""


def create_random_generator(settings=None):
    """Create a random generator from given settings.

    Parameters
    ----------
    settings : dict, optional
        This is the dict used for creating the random generator.
        Currently, we will actually just look for a seed value.

    Returns
    -------
    out : object like :py:class:`.RandomGenerator`
        The random generator created.

    """
    if settings is None:
        settings = {}

    rgen = settings.get('rgen', 'rgen')
    seed = settings.get('seed', 0)
    logger.debug('Seed for random generator: %s %d', rgen, seed)

    class_map = {
        'rgen': RandomGenerator,
        'rgen-borg': RandomGeneratorBorg,
        'mock': MockRandomGenerator,
        'mock-borg': MockRandomGeneratorBorg
    }
    rgen_class = class_map.get(rgen, RandomGenerator)
    rgen = rgen_class(seed=seed)

    if 'state' in settings:
        rgen.set_state(settings)
        rgen.status = 'restarted'
    else:
        rgen.status = 'new rgen'

    return rgen



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test module for flux analysis."""
import logging
import unittest
import warnings
import os
import pickle
import numpy as np
from pyretis.analysis import analyse_md_flux
from pyretis.analysis.flux_analysis import analyse_flux, find_crossings
from pyretis.inout.formats.cross import CrossFile
from pyretis.inout.formats.order import OrderFile
from pyretis.inout.settings import SECTIONS

logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


class FluxTest(unittest.TestCase):
    """Test that we can analyse for initial flux."""

    def test_flux_analysis(self):
        """Test the flux analysis."""
        filename = os.path.join(HERE, 'cross.txt')
        data = None
        with CrossFile(filename, 'r') as crossfile:
            data = crossfile.load()
        settings = {
            'simulation': {'endcycle': 250000, 'interfaces': [-0.9]},
            'engine': {'timestep': 0.002},
            'particles': {'npart': 1},
            'system': {'dimensions': 1,
                       'temperature': 666,
                       'beta': 1},
        }
        settings['analysis'] = SECTIONS['analysis']
        # What if empty?
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            results_all = analyse_md_flux(crossdata=[],
                                          energydata=[],
                                          orderdata=np.zeros(shape=(5, 2)),
                                          settings=settings)

        results = results_all['flux']
        self.assertFalse(results['eff_cross'])

        correct_file = os.path.join(HERE, 'flux-results.dat')
        for i in data:
            results = analyse_flux(i['data'], settings)
            break
        correct_data = {}
        with open(correct_file, 'rb') as infile:
            correct_data = pickle.load(infile)
        for key0 in ('eff_cross', 'ncross', 'neffcross',
                     'flux', 'runflux', 'interfaces', 'cross_time',
                     'neffc/nc', 'pMD', '1-p', 'teffMD', 'corrMD'):
            self.assertTrue(key0 in correct_data)
            self.assertTrue(key0 in results)
            for i, j in zip(correct_data[key0], results[key0]):
                self.assertTrue(np.allclose(i, j))
        self.assertAlmostEqual(correct_data['totalcycle'],
                               results['totalcycle'])
        for key, val in correct_data['times'].items():
            self.assertAlmostEqual(val, results['times'][key])
        for data1, data2 in zip(correct_data['errflux'],
                                results['errflux']):
            for i, j in zip(data1, data2):
                self.assertTrue(np.allclose(i, j))

    def test_flux_analysis_B(self):
        """Test the flux analysis considers also flux from state B."""
        fluxdata = [(592, 2, -1), (1171, 2, 1), (1642, 1, -1), (2279, 1, 1),
                    (2851, 3, -1), (3497, 2, 1), (3937, 2, -1), (4838, 2, -1),
                    (4967, 3, -1), (5288, 3, 1), (6666, 2, -1), (6727, 2, -1),
                    (6828, 3, -1), (6869, 3, 1), (6971, 2, -1), (9294, 3, 1)]
        settings = {
            'simulation': {'endcycle': 9999, 'interfaces': [-0.9, -0.8, -0.7]},
            'engine': {'timestep': 0.002},
            'particles': {'npart': 1},
            'system': {'dimensions': 1,
                       'temperature': 666,
                       'beta': 1},
        }
        settings['analysis'] = SECTIONS['analysis']
        results_all = analyse_flux(fluxdata, settings)
        self.assertEqual(results_all['eff_cross'][0][0], (637, 2279))
        self.assertEqual(results_all['teffMD'][0], 9999.0)
        self.assertEqual(results_all['corrMD'][0], 3.334111370456819)

    def test_crossing_calculation(self):
        """Test calculation of crossings from order parameter data."""
        filename = os.path.join(HERE, 'order-calculate-crossings.txt')
        data = None
        with OrderFile(filename, 'r') as orderfile:
            data = orderfile.load()
        orderp = next(data)['data'][:, 1]
        cross = find_crossings(orderp, [-0.9, -0.85, -0.70])

        filename_cross = os.path.join(HERE, 'cross-correct.txt')
        cross_correct = None
        with CrossFile(filename_cross, 'r') as crossfile:
            cross_correct = next(crossfile.load())['data']
        self.assertEqual(len(cross), len(cross_correct))
        for i, j in zip(cross, cross_correct):
            self.assertEqual(i[0], j[0])
            self.assertEqual(i[1], j[1] - 1)
            self.assertIn(i[2], ('+', '-'))
            if i[2] == '-':
                self.assertEqual(j[2], -1)
            elif i[2] == '+':
                self.assertEqual(j[2], 1)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A collection of simple position dependent potentials.

This module defines some potential functions which are useful
as simple models.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

DoubleWell (:py:class:`.DoubleWell`)
    This class defines a one-dimensional double well potential.

RectangularWell (:py:class:`.RectangularWell`)
    This class defines a one-dimensional rectangular well potential.
"""
import logging
import numpy as np
from pyretis.forcefield.potential import PotentialFunction
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['DoubleWell', 'RectangularWell']


class DoubleWell(PotentialFunction):
    r"""A 1D double well potential.

    This class defines a one-dimensional double well potential.
    The potential energy (:math:`V_\text{pot}`) is given by

    .. math::

       V_\text{pot} = a x^4 - b (x - c)^2

    where :math:`x` is the position and :math:`a`, :math:`b`
    and :math:`c` are parameters for the potential. These parameters
    are stored as attributes of the class. Typically, both :math:`a`
    and :math:`b` are positive quantities, however, we do not explicitly
    check that here.

    Attributes
    ----------
    params : dict
        Contains the parameters. The keys are:

        * `a`: The ``a`` parameter for the potential.
        * `b`: The ``b`` parameter for the potential.
        * `c`: The ``c`` parameter for the potential.

        These keys corresponds to the parameters in the potential,
        :math:`V_\text{pot} = a x^4 - b (x - c)^2`.

    """

    def __init__(self, a=1.0, b=1.0, c=0.0, desc='1D double well potential'):
        """Initialise the one dimensional double well potential.

        Parameters
        ----------
        a : float, optional
            Parameter for the potential.
        b : float, optional
            Parameter for the potential.
        c : float, optional
            Parameter for the potential.
        desc : string, optional
            Description of the force field.

        """
        super().__init__(dim=1, desc=desc)
        self.params = {'a': a, 'b': b, 'c': c}

    def potential(self, system):
        """Evaluate the potential for the one-dimensional double well.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential for. Here, we
            make use of the positions only.

        Returns
        -------
        out : float
            The potential energy.

        """
        pos = system.particles.pos
        v_pot = (self.params['a'] * pos**4 -
                 self.params['b'] * (pos - self.params['c'])**2)
        return v_pot.sum()

    def force(self, system):
        """Evaluate forces for the 1D double well potential.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential for. Here, we
            make use of the positions only.

        Returns
        -------
        out[0] : numpy.array
            The calculated force.
        out[1] : numpy.array
            The virial, currently not implemented for this potential!

        """
        pos = system.particles.pos
        forces = (-4.0*(self.params['a'] * pos**3) +
                  2.0*(self.params['b'] * (pos - self.params['c'])))
        virial = np.zeros((self.dim, self.dim))  # just return zeros here
        return forces, virial

    def potential_and_force(self, system):
        """Evaluate the potential and the force.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential for. Here, we
            make use of the positions only.

        Returns
        -------
        out[0] : float
            The potential energy as a float.
        out[1] : numpy.array
            The force as a numpy.array of the same shape as the
            positions in `particles.pos`.
        out[2] : numpy.array
            The virial, currently not implemented for this potential!

        """
        pos = system.particles.pos
        dist = pos - self.params['c']
        pos3 = pos**3
        v_pot = self.params['a'] * pos3 * pos - self.params['b'] * dist**2
        forces = (-4.0 * (self.params['a'] * pos3) +
                  2.0 * (self.params['b'] * dist))
        virial = np.zeros((self.dim, self.dim))  # just return zeros here
        return v_pot.sum(), forces, virial


class RectangularWell(PotentialFunction):
    r"""A 1D rectangular well potential.

    This class defines a one-dimensional rectangular well potential.
    The potential energy is zero within the potential well and infinite
    outside. The well is defined with a left and right boundary.

    Attributes
    ----------
    params : dict
        The parameters for the potential. The keys are:

        * `left`: Left boundary of the potential.
        * `right`: Right boundary of the potential.
        * `largenumber`: Value of potential outside the boundaries.

        It is possible to define left > right, however, a warning will
        be issued then.

    """

    def __init__(self, left=0.0, right=1.0, largenumber=float('inf'),
                 desc='1D Rectangular well potential'):
        """Initialise the one-dimensional rectangular well.

        Parameters
        ----------
        left : float, optional
            The left boundary of the potential.
        right : float, optional
            The right boundary of the potential.
        largenumber : float, optional
            The value of the potential outside (left, right).
        desc : string, optional
            Description of the force field.

        """
        super().__init__(dim=1, desc=desc)
        self.params = {'left': left, 'right': right,
                       'largenumber': largenumber}
        self.check_parameters()

    def check_parameters(self):
        """Check the consistency of the parameters.

        Returns
        -------
        out : None
            Returns `None` but might give a warning.

        """
        if self.params['left'] >= self.params['right']:
            msg = 'Setting left >= right in RectangularWell potential!'
            logger.warning(msg)

    def potential(self, system):
        """Evaluate the potential.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential for. Here, we
            make use of the positions only.

        Returns
        -------
        out : float
            The potential energy.

        """
        pos = system.particles.pos
        left = self.params['left']
        right = self.params['right']
        largenumber = self.params['largenumber']
        v_pot = np.where(np.logical_and(pos > left, pos < right),
                         0.0, largenumber)
        return v_pot.sum()  # pylint: disable=no-member



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test using the LAMMPS engine."""
import os
import sys
import colorama
from matplotlib import pyplot as plt
from pyretis.engines.lammps import LAMMPSEngine
from pyretis.inout import print_to_screen
from pyretis.inout.common import make_dirs
from pyretis.testing.systemhelp import create_system_ext
from pyretis.testing.helpers import clean_dir
from plotting import plot_compare, plot_xy


HERE = os.path.abspath(os.path.dirname(__file__))
STEPS = 50
SUBCYCLES = 2


def run_in_folder(system, engine, exe_dir, steps):
    """Run the engine in the given folder."""
    make_dirs(exe_dir)
    clean_dir(exe_dir)
    engine.exe_dir = exe_dir
    engine.integrate(system, steps, order_function=None, thermo='full')
    energy = engine.read_energies('pyretis_md')
    order = engine.read_order_parameters('order_pyretis_md.txt')
    return energy, order


def run_forward():
    """Use the LAMMPS engine to run a MD simulation forward in time."""
    print_to_screen(
        '\nTesting the LAMMPS engine by running forward.\n',
    )
    engine = LAMMPSEngine('lmp_serial', 'lammps_input', SUBCYCLES)
    print_to_screen(
        f'Running forward for {2 * STEPS} steps...',
        level='info',
    )
    # Create a dummy system:
    system = create_system_ext(pos=('system.data', 0))
    exe_dir = os.path.join(HERE, 'run-forward1')
    energy_f1, order_f1 = run_in_folder(system, engine, exe_dir, 2*STEPS)
    # Run forward, for the initial state for 100 steps:
    print_to_screen('-> Forward done! Resetting system.')
    print_to_screen(
        f'\nRunning forward for {STEPS} steps...',
        level='info',
    )
    system = create_system_ext(pos=('system.data', 0))
    exe_dir = os.path.join(HERE, 'run-forward2')
    energy_f2, order_f2 = run_in_folder(system, engine, exe_dir, STEPS)
    # Run for 50 more steps:
    print_to_screen('-> Forward done!')
    print_to_screen(
        f'\nContinuing run for {STEPS} steps...',
        level='info',
    )
    config = system.particles.get_pos()
    traj_file = os.path.join(exe_dir, config[0])
    system.particles.set_pos((traj_file, config[1]))
    exe_dir = os.path.join(HERE, 'run-forward3')
    energy_f3, order_f3 = run_in_folder(system, engine, exe_dir, STEPS)
    # Update step for part 3:
    energy_f3['Step'] += energy_f2['Step'][-1]
    print_to_screen('-> Forward done! Comparing energies:')

    data_sets = []

    for key in ['PotEng', 'KinEng', 'Press', 'Temp']:
        data_set = []
        for i, energy in enumerate((energy_f1, energy_f2, energy_f3)):
            if i == 0:
                lab = 'Full simulation'
            else:
                lab = f'Simulation, part {i}'
            data_set.append(
                (energy['Step'], energy[key], lab)
            )
        data_sets.append(data_set)

    plot_compare(data_sets, ['PotEng', 'KinEng', 'Press', 'Temp'])

    data_sets_xy = []
    for key in ('PotEng', 'KinEng', 'Press', 'Temp'):
        energy_parts = [i for i in energy_f2[key]]
        # Remove the data point that is common to part 1 and part 2:
        energy_parts.pop()
        # Add data points from part 2:
        energy_parts += [i for i in energy_f3[key]]
        data_sets_xy.append(
            (
                energy_f1[key], energy_parts,
                f'{key}, full simulation',
                f'{key}, part1 & part2'
            )
        )
    plot_xy(data_sets_xy)

    print_to_screen('-> Will also compare order parameters:')
    data_sets = [
        [
            (energy_f1['Step'], order_f1[:, 0], 'Full simulation'),
            (energy_f2['Step'], order_f2[:, 0], 'Simulation, part 1'),
            (energy_f3['Step'], order_f3[:, 0], 'Simulation, part 2'),
        ],
        [
            (energy_f1['Step'], order_f1[:, 1], 'Full simulation'),
            (energy_f2['Step'], order_f2[:, 1], 'Simulation, part 1'),
            (energy_f3['Step'], order_f3[:, 1], 'Simulation, part 2'),
        ],
    ]
    plot_compare(data_sets, ['OP1', 'OP2'])

    data_sets_xy = []
    for i in (0, 1):
        order_parts = [i for i in order_f2[:, i]]
        # Remove the data point that is common to part 1 and part 2:
        order_parts.pop()
        # Add data points from part 2:
        order_parts += [i for i in order_f3[:, i]]
        data_sets_xy.append(
            (
                order_f1[:, i], order_parts,
                f'OP{i + 1}, full simulation',
                f'OP{i + 1}, part1 & part2'
            )
        )
    plot_xy(data_sets_xy)


def run_forward_backward():
    """Use the LAMMPS engine to run a MD simulation forward in time."""
    print_to_screen(
        '\nTesting the LAMMPS engine by running forward and backward.\n',
    )
    # Set up for running forward:
    print_to_screen(
        f'-> Running forward for {STEPS} steps...',
        level='info',
    )
    engine = LAMMPSEngine('lmp_serial', 'lammps_input', SUBCYCLES)
    system = create_system_ext(pos=('system.data', 0))
    exe_dir = os.path.join(HERE, 'run-forward')
    energy_f, order_f = run_in_folder(system, engine, exe_dir, STEPS)
    print_to_screen('-> Forward done! Prepare backward run.')
    # Reverse the velocities and run:
    config = system.particles.get_pos()
    traj_file = os.path.join(exe_dir, config[0])
    print_to_screen('-> Setting configuration.')
    system.particles.set_pos((traj_file, config[1]))
    print_to_screen('-> Reversing velocities.')
    system.particles.set_vel(True)

    print_to_screen(f'\n-> Starting backward run ({STEPS}) steps!',
                    level='info')
    exe_dir = os.path.join(HERE, 'run-backward')
    energy_b, order_b = run_in_folder(system, engine, exe_dir, STEPS)
    print_to_screen('\n-> Backward done! Will compare energies:')

    step = energy_b['Step'] - energy_b['Step'][0]
    data_sets = [
        [
            (energy_f['Step'], energy_f['PotEng'],
             'Forward'),
            (step, energy_b['PotEng'][::-1], 'Backward'),
        ],
        [
            (energy_f['Step'], energy_f['KinEng'],
             'Forward'),
            (step, energy_b['KinEng'][::-1], 'Backward'),
        ],
        [
            (energy_f['Step'], energy_f['Press'],
             'Forward'),
            (step, energy_b['Press'][::-1], 'Backward'),
        ],
        [
            (energy_f['Step'], energy_f['Temp'],
             'Forward'),
            (step, energy_b['Temp'][::-1], 'Backward'),
        ],
    ]
    plot_compare(data_sets, ['PotEng', 'KinEng', 'Press', 'Temp'])

    data_sets_xy = [
        (energy_f['PotEng'], energy_b['PotEng'][::-1],
         'PotEng, forward', 'PotEng, backward'),
        (energy_f['KinEng'], energy_b['KinEng'][::-1],
         'KinEng, forward', 'KinEng, backward'),
        (energy_f['Press'], energy_b['Press'][::-1],
         'Press, forward', 'Press, backward'),
        (energy_f['Temp'], energy_b['Temp'][::-1],
         'Temp, forward', 'Temp, backward'),
    ]
    plot_xy(data_sets_xy)

    print_to_screen('\n-> Will also compare order parameters:')
    print_to_screen('NOTE: OP2 IS HERE MULTIPLIED WITH "-1"', level='warning')
    print_to_screen('NOTE: THIS, SINCE IT IS VELOCITY DEPENDENT!',
                    level='warning')
    data_sets = [
        [
            (step, order_f[:, 0], 'Forward'),
            (step, order_b[:, 0][::-1], 'Backward'),
        ],
        [
            (step, order_f[:, 1], 'Forward'),
            (step, -1 * order_b[:, 1][::-1], 'Backward'),
        ],
    ]
    plot_compare(data_sets, ['OP1', 'OP2'])

    data_sets_xy = [
        (order_f[:, 0], order_b[:, 0][::-1], 'OP1, forward', 'OP1, backward'),
        (order_f[:, 1], -1 * order_b[:, 1][::-1],
         'OP2, forward', 'OP2, backward'),
    ]
    plot_xy(data_sets_xy)


def main(plot=False):
    """Run the comparisons."""
    run_forward()
    run_forward_backward()
    if plot:
        plt.show()


if __name__ == '__main__':
    colorama.init(autoreset=True)
    main(plot=len(sys.argv) >= 2)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains examples for order parameters.

This file is distributed as part of the documentation of PyRETIS.

Note: mdtraj needs to be installed on top of the regular PyRETIS installation.

"""
import mdtraj
from pyretis.orderparameter import OrderParameter


def some_function_to_obtain_order(_):
    """A always right function test."""
    return 42


class PlaneDistanceX(OrderParameter):
    """A positional order parameter.

    This class defines a very simple order parameter which is
    the distance from a plane for a given particle.
    """

    def __init__(self, index, plane_position):
        """Initialise the order parameter.

        Parameters
        ----------
        index : integer
            Selects the particle to use.

        plane_position : float
            The location of the plane, along the x-axis.
        """
        txt = f'Distance from particle {index} to a plane at {plane_position}'
        super().__init__(description=txt)
        self.index = index
        self.plane_position = plane_position

    def calculate(self, system):
        """Calculate the order parameter."""
        filename = system.particles.config[0]
        traj = mdtraj.load(filename)
        return some_function_to_obtain_order(traj)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""PyRETIS - A simulation package for rare event simulations.

PyRETIS - A simulation package for rare event simulations.
Copyright (c) 2026, PyRETIS Development Team

This file is part of PyRETIS.

PyRETIS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 2.1 of the License, or
(at your option) any later version.

PyRETIS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyRETIS. If not, see <http://www.gnu.org/licenses/>
"""
import ast
from codecs import open as openc
import pathlib
from setuptools import setup, find_namespace_packages

FULL_VERSION = '3.0.1'  # Automatically set by setup_version.py


def get_long_description():
    """Return the contents of README.rst as a string."""
    here = pathlib.Path(__file__).absolute().parent
    long_description = ''
    with openc(here.joinpath('README.rst'), encoding='utf-8') as fileh:
        long_description = fileh.read()
    return long_description


def get_version():
    """Return the version from version.py as a string."""
    here = pathlib.Path(__file__).absolute().parent
    filename = here.joinpath('pyretis', 'version.py')
    with openc(filename, encoding='utf-8') as fileh:
        for lines in fileh:
            if lines.startswith('FULL_VERSION ='):
                version = ast.literal_eval(lines.split('=')[1].strip())
                return version
    return FULL_VERSION


def get_requirements(docs=False):
    """Read requirements.txt and return a list of requirements."""
    here = pathlib.Path(__file__).absolute().parent
    requirements = []
    doc_package = ('sphinx', 'sphinx_bootstrap_theme')
    filename = here.joinpath('requirements.txt')
    with openc(filename, encoding='utf-8') as fileh:
        for lines in fileh:
            if '>=' in lines:
                package = lines.split('>=')[1].strip()
            elif '<' in lines:
                package = lines.split('<')[1].strip()
            if not docs and package in doc_package:
                continue
            requirements.append(lines.strip())
    return requirements


setup(
    name='pyretis',
    version=get_version(),
    description='A simulation package for rare events',
    long_description=get_long_description(),
    url='http://www.pyretis.org',
    author='PyRETIS Development Team',
    author_email='pyretis@pyretis.org',
    license='LGPLv2.1+',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        ('License :: OSI Approved :: '
         'GNU Lesser General Public License v2 or later (LGPLv2+)'),
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    keywords='rare-events md mc tps simulation tis retis',
    packages=find_namespace_packages(exclude=['docs']),
    package_data={
        'pyretis': [
            'pyretis.mplstyle',
            'pyretis/inout/report/templates/*',
            'pyretis/pyvisa/pyvisaGUI.ui'
        ]
    },
    include_package_data=True,
    install_requires=get_requirements(docs=False),
    entry_points={
        'console_scripts': [
            'pyretisrun = pyretis.bin.pyretisrun:entry_point',
            'pyretisanalyse = pyretis.bin.pyretisanalyse:entry_point',
            'pyvisa = pyretis.bin.pyvisa:entry_point'
        ]
    },
)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the recalculate orderparameter tools in pyretis.tools"""
from contextlib import contextmanager
from io import StringIO
import logging
import os
import tarfile
from tempfile import NamedTemporaryFile
import unittest
from unittest.mock import patch
import numpy as np
from pyretis.tools.recalculate_order import (
    recalculate_from_frame,
    recalculate_from_trj,
    recalculate_order,
    get_traj_files,
)

from pyretis.orderparameter import OrderParameter


logging.disable(logging.CRITICAL)


@contextmanager
def turn_on_logging():
    """Turn on logging so that tests can detect it."""
    logging.disable(logging.INFO)
    try:
        yield
    finally:
        logging.disable(logging.CRITICAL)


CORRECT_TRI = [
    [1.2000000476837158, 1.100000023841858, 0.9526299834251404,
     0.0, 0.0, 0.0, 0.0, 0.550000011920929, 0.0],
    [1.3190346956253052, 1.1973484754562378, 1.0101481676101685,
     0.0, 0.0, 0.0, 0.0, 0.6045578122138977, 0.0],
    [1.445665717124939, 1.2725739479064941, 1.043269157409668,
     0.0, 0.0, 0.0, 0.0, 0.6625972390174866, 0.0],
    [1.549871802330017, 1.3433609008789062, 1.0741169452667236,
     0.0, 0.0, 0.0, 0.0, 0.7103575468063354, 0.0],
    [1.6492804288864136, 1.3989951610565186, 1.0993438959121704,
     0.0, 0.0, 0.0, 0.0, 0.7559200525283813, 0.0],
    [1.7235912084579468, 1.4501303434371948, 1.1253700256347656,
     0.0, 0.0, 0.0, 0.0, 0.7899801731109619, 0.0],
    [1.8234614133834839, 1.5086286067962646, 1.1493875980377197,
     0.0, 0.0, 0.0, 0.0, 0.8357531428337097, 0.0],
    [1.9387280941009521, 1.5735180377960205, 1.1708111763000488,
     0.0, 0.0, 0.0, 0.0, 0.888584315776825, 0.0],
    [2.0414657592773438, 1.6291977167129517, 1.1705245971679688,
     0.0, 0.0, 0.0, 0.0, 0.9356740117073059, 0.0],
    [2.138443946838379, 1.6822302341461182, 1.191916584968567,
     0.0, 0.0, 0.0, 0.0, 0.9801228642463684, 0.0],
    [2.234323024749756, 1.7414289712905884, 1.219726800918579,
     0.0, 0.0, 0.0, 0.0, 1.0240683555603027, 0.0],
]


CORRECT_XYZ = [
    [-1.0912E-05, 1.85E-07, -2.310E-06],
    [-2.1178E-05, 6.2E-08, -4.785E-06],
    [-3.1073E-05, -3.8E-07, -7.123E-06],
    [-4.1081E-05, -1.026E-06, -9.074E-06],
    [-5.1297E-05, -1.657E-06, -1.0489E-05],
    [-6.1028E-05, -2.03E-06, -1.1326E-05],
]


HERE = os.path.abspath(os.path.dirname(__file__))


class OrderBox(OrderParameter):
    """This order parameter is just the box."""

    def __init__(self):
        super().__init__(description='Box order parameter')

    def calculate(self, system):
        return system.box.cell


class Xpos1partExt(OrderParameter):
    """This gives out the X position of the first particle."""

    def __init__(self):
        super().__init__(description='X pos particle 1')

    def calculate(self, system):
        return system.particles.pos[0][0]


class Velocity(OrderParameter):
    """This order parameter is the sum of components of the velocity."""

    def __init__(self):
        super().__init__(description='Box order parameter')

    def calculate(self, system):
        return system.particles.vel.sum(axis=0)


class TestRecalculateOrder(unittest.TestCase):
    """Test the calculation of order parameters from trajectory files."""

    def test_rectangular_box(self):
        """Test recalculate for a rectangular box."""
        orderf = OrderBox()
        box = np.array([2.7200000286102295]*3)
        trr = os.path.join(HERE, '4water.trr')
        with patch('sys.stdout', new=StringIO()):
            for i in recalculate_order(orderf, trr, {}):
                self.assertTrue(np.allclose(i, box))

    def test_novel(self):
        """Test recalculate for a rectangular box."""
        trr = os.path.join(HERE, 'traj-novel.trr')
        with turn_on_logging(), patch('sys.stdout', new=StringIO()):
            with self.assertLogs('pyretis.tools.recalculate_order',
                                 level='WARNING'):

                for order in recalculate_order(Velocity(), trr, {}):
                    for i in order:
                        self.assertAlmostEqual(i, 0.0)

    def test_tri_box(self):
        """Test recalculate for a triclinic box."""
        orderf = OrderBox()
        trr = os.path.join(HERE, '4water_tri_dyn.trr')
        orderp = []
        with patch('sys.stdout', new=StringIO()):
            for i, j in zip(recalculate_order(orderf, trr, {}), CORRECT_TRI):
                orderp.append(i)
                self.assertTrue(np.allclose(i, j))
            self.assertEqual(len(orderp), len(CORRECT_TRI))
            for i, j in zip(
                    recalculate_order(orderf, trr, {'reverse': True}),
                    orderp[::-1]):
                self.assertTrue(np.allclose(i, j))

    def test_sub_selection(self):
        """Test when we specify a min/max index to consider."""
        orderf = OrderBox()
        trr = os.path.join(HERE, '4water_tri_dyn.trr')
        idx = 3
        with patch('sys.stdout', new=StringIO()):
            orderp = [i for i in recalculate_order(orderf, trr,
                                                   {'maxidx': idx})]
            for i, j in zip(orderp, CORRECT_TRI[:idx+1]):
                self.assertTrue(np.allclose(i, j))
        with patch('sys.stdout', new=StringIO()):
            orderp = [i for i in recalculate_order(orderf, trr,
                                                   {'minidx': idx})]
            for i, j in zip(orderp, CORRECT_TRI[idx:]):
                self.assertTrue(np.allclose(i, j))

    def test_recalculate_gro(self):
        """Recalculate from .gro file."""
        orderf = OrderBox()
        tar = tarfile.open(
            os.path.join(HERE, '4water_tri_dyn.gro.tgz'), 'r:gz'
        )
        members = sorted([i.name for i in tar.getmembers()])
        i = 0
        with patch('sys.stdout', new=StringIO()):
            for member in members:
                gro = tar.extractfile(member)
                if gro is not None:
                    grofile = NamedTemporaryFile(suffix='.gro')
                    with open(grofile.name, 'wb') as output:
                        output.write(gro.read())
                    box = recalculate_order(orderf, grofile.name, {})
                    for j, k in zip(box[0], CORRECT_TRI[i]):
                        self.assertAlmostEqual(j, k, places=5)
                    i += 1
        # Do some additional tests:
        with patch('sys.stdout', new=StringIO()):
            gro = tar.extractfile(members[1])
            grofile = NamedTemporaryFile(suffix='.gro')
            with open(grofile.name, 'wb') as output:
                output.write(gro.read())
            vel1, = recalculate_from_frame(Velocity(), grofile.name,
                                           {'ext': '.gro', 'reverse': False})
            vel2, = recalculate_from_frame(Velocity(), grofile.name,
                                           {'ext': '.gro', 'reverse': True})
            self.assertTrue(np.allclose(vel1, -vel2))
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError):
                recalculate_from_frame(Velocity(), 'something.txt',
                                       {'ext': '.txt'})

    def test_recalculate_g96(self):
        """Recalculate from .g96."""
        orderf = OrderBox()
        tar = tarfile.open(
            os.path.join(HERE, '4water_tri_dyn.g96.tgz'), 'r:gz'
        )
        members = sorted([i.name for i in tar.getmembers()])
        i = 0
        with patch('sys.stdout', new=StringIO()):
            for member in members:
                g96 = tar.extractfile(member)
                if g96 is not None:
                    g96file = NamedTemporaryFile(suffix='.g96')
                    with open(g96file.name, 'wb') as output:
                        output.write(g96.read())
                    box = recalculate_order(orderf, g96file.name, {})
                    for j, k in zip(box[0], CORRECT_TRI[i]):
                        self.assertAlmostEqual(j, k, places=5)
                    i += 1

    def test_recalculate_xyz(self):
        """Recalculate from .xyz."""
        orderf = Velocity()
        filename = os.path.join(HERE, 'traj.xyz')
        with patch('sys.stdout', new=StringIO()):
            order = [i for i in recalculate_order(orderf, filename, {})]
            order_r = [
                i for i in recalculate_order(
                    orderf, filename, {'reverse': True}
                )
            ]
            self.assertEqual(len(order), len(CORRECT_XYZ))
            self.assertEqual(len(order), len(order_r))
            for vel1, vel2, vel3 in zip(order, order_r[::-1], CORRECT_XYZ):
                for i, j, k in zip(vel1, vel2, vel3):
                    self.assertAlmostEqual(i, -j, places=7)
                    self.assertAlmostEqual(i, k, places=7)
        # Test with subsection:
        idx = 2
        with patch('sys.stdout', new=StringIO()):
            order = [
                i for i in recalculate_order(orderf, filename, {'maxidx': idx})
            ]
            for i, j in zip(order, CORRECT_XYZ[:idx+1]):
                self.assertTrue(np.allclose(i, j))
        with patch('sys.stdout', new=StringIO()):
            order = [
                i for i in recalculate_order(orderf, filename, {'minidx': idx})
            ]
            for i, j in zip(order, CORRECT_XYZ[idx:]):
                self.assertTrue(np.allclose(i, j))
        # Test with missing box:
        filename = os.path.join(HERE, 'traj-no-box.xyz')
        orderf = OrderBox()
        with patch('sys.stdout', new=StringIO()):
            order = [
                i for i in recalculate_order(orderf, filename,
                                             {'box': [6., 6., 6.]})
            ]
            for orderi in order:
                self.assertEqual(len(orderi), 3)
                for i in orderi:
                    self.assertAlmostEqual(i, 6.)

    def test_get_traj_files(self):
        """Test the method for getting traj files."""
        traj = os.path.join(HERE, 'traj-ext.txt')
        correct = {
            'trajB.trr': {'minidx': 0, 'maxidx': 6, 'reverse': True},
            'trajF.trr': {'minidx': 0, 'maxidx': 7, 'reverse': False},
        }
        files = get_traj_files(traj)
        for key, val in correct.items():
            self.assertIn(key, files)
            self.assertEqual(val, files[key])

    def test_use_exteral(self):
        """Test the method for load and recalculate op from ext trjs."""
        orderf = Xpos1partExt()
        sim_file = os.path.join(HERE, '2water')

        ordercheck = [0.1259, 0.1257, 0.1255, 0.1253, 0.1251, 0.1249]

        with patch('sys.stdout', new=StringIO()):
            order = [i for i in recalculate_order(
                order_parameter=orderf,
                traj_file=sim_file+'.trr',
                options={'top': sim_file+'.gro'})]

        for orderi, orderj in zip(order, ordercheck):
            self.assertAlmostEqual(orderi, orderj, 3, 0.1)

        with patch('sys.stdout', new=StringIO()):
            order = [i for i in recalculate_from_trj(
                orderf, sim_file + '.trr',
                {'top': sim_file + '.gro'})]

        for orderi, orderj in zip(order, ordercheck):
            self.assertAlmostEqual(orderi, orderj, 3, 0.1)

        one_op = recalculate_from_trj(orderf, sim_file + '.trr',
                                      {'top': sim_file + '.gro',
                                       'idx': 3})

        with patch('sys.stdout', new=StringIO()):
            self.assertAlmostEqual(next(one_op), ordercheck[3], 3, 0.1)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Time the FORTRAN implementation of the Lennard-Jones potential.

This timing is simply done by evaluating the Lennard-Jones forces
(and potential) for different system sizes.
"""
# pylint: disable=invalid-name
import timeit
import numpy as np
from pyretis.core import System, create_box, Particles
from pyretis.core.units import create_conversion_factors
from pyretis.tools import generate_lattice
from ljpotentialf import PairLennardJonesCutF


def set_up_initial_state(nlattice=5):
    """Create particles for the test.

    This will set up a 3D lattice with 4*nlattice**3 particles.
    """
    create_conversion_factors('lj')
    lattice, size = generate_lattice('fcc', [nlattice] * 3, density=0.9)
    npart = len(lattice)
    lattice += np.random.randn(npart, 3) * 0.05
    size = np.array(size)
    box = create_box(low=size[:, 0], high=size[:, 1],
                     periodic=[True, True, True])
    sys = System(temperature=1.0, units='lj', box=box)
    sys.particles = Particles(dim=3)
    for pos in lattice:
        sys.add_particle(name='Ar', pos=pos, mass=1.0, ptype=0)
    msg = 'Created lattice with {} atoms.'
    print(msg.format(sys.particles.npart))
    return sys


def test_wrapper(func, *args, **kwargs):
    """A simple wrapper for calling functions."""
    def wrapped():
        return func(*args, **kwargs)
    return wrapped


def test_function(function, system, repeat=3, number=5):
    """Run the test for a function."""
    print(f'Testing function: {function.__name__}')
    wrapped = test_wrapper(function, system)
    res = timeit.repeat(wrapped, repeat=repeat, number=number)
    best = min(res) / float(number)
    avg = np.average([resi / float(number) for resi in res])
    std = np.std([resi / float(number) for resi in res])
    print(f'Best: {best}')
    print(f'Average: {avg} +- {std}')
    return best, avg, std


if __name__ == '__main__':
    parameters = {0: {'sigma': 1.0, 'epsilon': 1.0, 'rcut': 2.5}}
    # set up potentials:
    potential = PairLennardJonesCutF(dim=3, shift=True, mixing='geometric')
    potential.set_parameters(parameters)

    results = []

    for i in range(3, 16):
        system = set_up_initial_state(nlattice=i)
        print('Testing FORTRAN implementation')
        time1 = test_function(potential.potential_and_force,
                              system,
                              number=10, repeat=3)
        results.append((system.particles.npart, time1[0], time1[1], time1[2]))
    results = np.array(results)
    np.savetxt('timings.txt', results, fmt='%i %.9e %.9e %.9e',
               header='N best avg std')



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Some methods for generating initial lattice structures.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

generate_lattice (:py:func:`.generate_lattice`)
    Generate points on a simple lattice.

Examples
--------
>>> from pyretis.tools.lattice import generate_lattice
>>> xyz, size = generate_lattice('diamond', [1, 1, 1], lcon=1)

"""
import logging
import itertools
import numpy as np
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

__all__ = ['generate_lattice']


UNIT_CELL = {'sc': np.array([[0.0, 0.0, 0.0]]),
             '1d': np.array([[0.0]]),
             'sq': np.array([[0.0, 0.0]]),
             'sq2': np.array([[0.0, 0.0], [0.5, 0.5]]),
             'bcc': np.array([[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]),
             'fcc': np.array([[0.0, 0.0, 0.0], [0.5, 0.5, 0.0],
                              [0.0, 0.5, 0.5], [0.5, 0.0, 0.5]]),
             'hcp': np.array([[0.0, 0.0, 0.0], [0.5, 0.5, 0.0],
                              [0.5, 5.0/6.0, 0.5], [0.0, 1.0/3.0, 0.5]]),
             'diamond': np.array([[0.0, 0.0, 0.0], [0.0, 0.5, 0.5],
                                  [0.5, 0.0, 0.5], [0.5, 0.5, 0.0],
                                  [0.25, 0.25, 0.25], [0.25, 0.75, 0.75],
                                  [0.75, 0.25, 0.75], [0.75, 0.75, 0.25]])}


def generate_lattice(lattice, repeat=None, lcon=None, density=None):
    """Generate points on a simple lattice.

    The lattice is one of the defined keys in the global variable
    `UNIT_CELL`. This lattice will be repeated a number of times.
    The lattice spacing can be given explicitly, or it can be given
    implicitly by the number density.

    Parameters
    ----------
    lattice : string
        Select the kind of lattice. The following options are currently
        defined in `UNIT_CELL`:

         * `1d` : 1D lattice
         * `sc` : Simple cubic lattice.
         * `sq` : Square lattice (2D) with one atom in the unit cell.
         * `sq2` : Square lattice with two atoms in the unit cell.
         * `bcc` : Body-centred cubic lattice.
         * `fcc` : Face-centred cubic lattice.
         * `hcp` : Hexagonal close-packed lattice.
         * `diamond` : Diamond structure.

    repeat : list of integers, optional.
        How many time the cell is replicated.
    lcon : float, optional
        The lattice constant.
    density : float, optional
        The desired density. If this is given, `lcon` is calculated.
        Note that density will be interpreted as given in internal
        units.

    Returns
    -------
    positions : numpy.array
        The lattice positions.
    size : list of floats
        The corresponding size(s), can be used to define a simulation
        box.

    """
    try:
        unit_cell = UNIT_CELL[lattice.lower()]
    except KeyError as err:
        msg = [f'Unknown lattice "{lattice}" requested!']
        msg += [f'Input lattice should be a string in {UNIT_CELL.keys()}']
        raise ValueError('\n'.join(msg)) from err
    except AttributeError as err:
        msgtxt = f'Input lattice should be a string in {UNIT_CELL.keys()}'
        raise ValueError(msgtxt) from err
    ndim = len(unit_cell[0])
    npart = len(unit_cell)
    if density is not None:
        lcon = (npart / density)**(1.0 / float(ndim))
    if repeat is None and density is None and lcon is None:
        logger.debug('The construction of the simulation '
                     'box is attempted with insufficient '
                     'information. Please check')
        repeat = [1]*ndim
        lcon = 1
    elif lcon is None:
        msgtxt = 'Could not determine lattice constant!'
        raise ValueError(msgtxt)
    elif len(repeat) < ndim:
        msgtxt = 'To few "repeat" values given: Expected {} but got {}.'
        raise ValueError(msgtxt.format(ndim, len(repeat)))
    positions = []
    for i in itertools.product(*[range(nri) for nri in repeat[:ndim]]):
        pos = lcon * (np.array(i) + unit_cell)
        positions.extend(pos)
    size = [[0.0, i * lcon] for i in repeat[:ndim]]
    positions = np.array(positions)
    return positions, size



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module handles the creation of simulations from settings.

The different simulations are defined as objects which inherit
from the base :py:class:`.Simulation` class defined in
:py:mod:`pyretis.simulation.simulation`. Here, we are treating
each simulation with a special case.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

create_simulation (:py:func:`.create_simulation`)
    Method for creating a simulation object from settings.

create_ensemble (:py:func:`.create_ensemble`)
    Method for creating an ensemble dictionary from settings.

create_ensembles (:py:func:`.create_ensembles`)
    Method for creating a list of ensemble from settings.

create_nve_simulation (:py:func:`.create_mve_simulation`)
    Method for creating a nve simulation object from settings.

create_mdflux_simulation (:py:func:`.create_mdflux_simulation`)
    Method for creating a mdflux simulation object from settings.

create_umbrellaw_simulation (:py:func:`.create_umbrellaw_simulation`)
    Method for creating a umbrellaw simulation object from settings.

create_tis_simulation (:py:func:`.create_tis_simulation`)
    Method for creating a tis simulation object from settings.

create_retis_simulation (:py:func:`.create_retis_simulation`)
    Method for creating a retis simulation object from settings.

prepare_system (:py:func:`.prepare_system`)
    Method for creating a system object from settings.

prepare_engine (:py:func:`.prepare_engine`)
    Method for creating an engine object from settings.

"""
import logging
import os
from pyretis.core.pathensemble import get_path_ensemble_class
from pyretis.setup.createforcefield import create_force_field
from pyretis.inout import print_to_screen
from pyretis.inout.settings import (add_default_settings,
                                    add_specific_default_settings,
                                    settings_from_restart)
from pyretis.setup.common import create_orderparameter
from pyretis.core.units import units_from_settings
from pyretis.core.random_gen import create_random_generator
from pyretis.simulation.md_simulation import (
    SimulationMD,
    SimulationNVE,
    SimulationMDFlux,
)
from pyretis.simulation.mc_simulation import UmbrellaWindowSimulation
from pyretis.simulation.path_simulation import (
    SimulationTIS,
    SimulationRETIS
)
from pyretis.setup import (
    create_system,
    create_engine
)
from pyretis.inout.checker import (
    check_engine,
    check_ensemble,
)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['create_simulation', 'create_ensemble', 'create_ensembles',
           'create_nve_simulation', 'create_mdflux_simulation',
           'create_umbrellaw_simulation', 'create_tis_simulation',
           'create_retis_simulation', 'prepare_system', 'prepare_engine']


def create_ensemble(settings):
    """Create the path ensemble from (ensemble) simulation settings.

    Parameters
    ----------
    settings : dict
        This dict contains the settings needed to create the path
        ensemble.

    Returns
    -------
    ensemble : dict of
        objects that contains all the information needed in the ensemble.

    """
    i_ens = settings['tis']['ensemble_number']

    logtxt = f'\nCREATING  ENSEMBLE  {i_ens}\n====================='
    print_to_screen(logtxt, level='message')
    logger.info(logtxt)

    rgen_ens = create_random_generator(settings['tis'])
    rgen_path = create_random_generator(settings['system'])

    system = prepare_system(settings)
    engine = prepare_engine(settings)
    klass = get_path_ensemble_class(settings['particles']['type'])
    interfaces = settings['simulation']['interfaces']
    exe_dir = settings['simulation'].get('exe_path', os.path.abspath('.'))
    path_ensemble = klass(i_ens, interfaces, rgen=rgen_path, exe_dir=exe_dir)

    # for PPRETIS / PPTIS: correct the starting condition of the paths
    if settings['simulation']['task'] in ['pptis', 'repptis']:
        change_start_cond = True
        # but skip this change if we are dealing with the [0-] ensemble
        if i_ens == 0:  # the first ensemble
            if settings['simulation'].get('flux', False) or \
                    settings['simulation'].get('zero_left', False):
                change_start_cond = False

        if change_start_cond:
            # adapt the starting condition of the paths
            logger.info("adapted start_condition to R/L in ensemble %s", i_ens)
            print('booger')
            path_ensemble.start_condition = ['R', 'L']
            path_ensemble.must_cross_M = True  # used in the shooting move

    order_function = create_orderparameter(settings)
    # code for computing the permeability
    if i_ens == 0 and settings['simulation'].get("permeability", False):
        # the starting condition in [0-'] can be R or L
        path_ensemble.start_condition = ['R', 'L']
        # check mirror move settings
        if hasattr(order_function, 'mirror_pos'):
            offset = getattr(order_function, 'offset', 0)
            moved_interfaces = [i - offset for i in interfaces]
            left, right = moved_interfaces[0], moved_interfaces[2]
            correct_mirror = (left+right)/2.
            if abs(order_function.mirror_pos-correct_mirror) > 1E-5:
                msg = "Order function should have a mirror at "
                msg += f"{correct_mirror}, found one at "
                msg += f"{order_function.mirror_pos} instead."
                raise ValueError(msg)
    # Add check to see if mirror makes sense
    engine.can_use_order_function(order_function)
    ensemble = {'engine': engine,
                'system': system,
                'order_function': order_function,
                'interfaces': interfaces,
                'exe_path': exe_dir,
                'path_ensemble': path_ensemble,
                'rgen': rgen_ens}
    ensemble['system'].order = engine.calculate_order(ensemble)

    return ensemble


def create_ensembles(settings):
    """Create a list of dictionary from ensembles simulation settings.

    Parameters
    ----------
    settings : dict
        This dict contains the settings needed to create the path
        ensemble.

    Returns
    -------
    ensembles : list of dicts
        List of ensembles.

    """
    # Example:
    # ensembles, assuming len(interfaces) = 3
    # (RE)TIS:   flux=False, zero_ensemble=False  [1+]
    # (RE)TIS:   flux=False, zero_ensemble=True   [0+], [1+]
    # (RE)TIS:   flux=True,  zero_ensemble=True   [0-], [0+], [1+]
    # /      :   flux=True,  zero_ensemble=False  doesn't make sense
    # so number of ensembles can be 1, 2, or 3
    ensembles = []
    intf = settings['simulation']['interfaces']
    # for PPTIS or REPPTIS: memory
    mem = settings.get(settings['simulation']['task'], {}).get('memory', 1)

    j_ens = 0
    # add [0-] if it exists
    if settings['simulation'].get('flux', False) or \
            settings['simulation'].get('zero_left', False):
        reactant, middle, product = float('-inf'), intf[0], intf[0]
        if settings['simulation'].get('zero_left', False):
            reactant = settings['simulation']['zero_left']
            if settings['simulation'].get('permeability', False):
                middle = (settings['simulation']['zero_left'] + intf[0]) / 2
        settings['ensemble'][j_ens]['simulation']['interfaces'] =\
            [reactant, middle, product]
        j_ens += 1

    # add [0+] if it exists
    if settings['simulation'].get('zero_ensemble', True):
        reactant, middle, product = intf[0], intf[0], intf[-1]

        # for PPTIS or REPPTIS: overwrite 'product' interface
        if settings['simulation']['task'] in ['pptis', 'repptis']:
            # mem is an integer, DEFAULT is 1
            # intf cannot be more to the right than intf[-1]
            product = intf[min(mem, len(intf)-1)]

        settings['ensemble'][j_ens]['simulation']['interfaces'] = \
            [reactant, middle, product]
        j_ens += 1

    # j_ens is the number of ensembles that is skipped

    # set interfaces and set detect for [1+], [2+], ...
    reactant, product = intf[0], intf[-1]
    for i, i_ens in enumerate(range(j_ens, len(settings['ensemble']))):
        middle = intf[i + 1]    # the lambda_i interface

        # for PPTIS / REPPTIS: overwrite 'reactant' and 'product' intf
        if settings['simulation']['task'] in ['pptis', 'repptis']:
            # mem is an integer, DEFAULT is 1
            # intf cannot be more to the left than intf[0]
            reactant = intf[max(0, i + 1 - mem)]
            # intf cannot be more to the right than intf[-1]
            product = intf[min(i + 1 + mem, len(intf)-1)]
            # middle is not changed, actually

        settings['ensemble'][i_ens]['simulation']['interfaces'] = \
            [reactant, middle, product]
        settings['ensemble'][i_ens]['tis']['detect'] = intf[i + 2]  # next intf

    # create all ensembles
    for i in range(len(settings['ensemble'])):
        ensembles.append(create_ensemble(settings['ensemble'][i]))

    return ensembles


def create_nve_simulation(settings):
    """Set up and create a NVE simulation.

    Parameters
    ----------
    settings : dict
        The settings needed to set up the simulation.

    Returns
    -------
    SimulationNVE : object like :py:class:`.SimulationNVE`
        The object representing the simulation to run.

    """
    ensemble = {'engine': prepare_engine(settings),
                'system': prepare_system(settings),
                'rgen': create_random_generator(settings['simulation'])}

    key_check('steps', settings)

    controls = {'steps': settings['simulation']['steps'],
                'startcycle': settings['simulation'].get('startcycle', 0)}

    return SimulationNVE(ensemble, settings, controls)


def create_md_simulation(settings):
    """Set up and create a generic MD simulation.

    Parameters
    ----------
    settings : dict
        The settings needed to set up the simulation.

    Returns
    -------
    out : object like :py:class:`.SimulationMD`
        The object representing the simulation to run.

    """
    ensemble = {'engine': prepare_engine(settings),
                'system': prepare_system(settings),
                'order_function': create_orderparameter(settings)}

    key_check('steps', settings)

    controls = {'steps': settings['simulation']['steps'],
                'startcycle': settings['simulation'].get('startcycle', 0)}

    return SimulationMD(ensemble, settings, controls)


def create_mdflux_simulation(settings):
    """Set up and create a MD FLUX simulation.

    Parameters
    ----------
    settings : dict
        The settings needed to set up the simulation.

    Returns
    -------
    SimulationMDFlux : object like :py:class:`.SimulationMDFlux`
        The object representing the simulation to run.

    """
    engine = prepare_engine(settings)
    order_function = create_orderparameter(settings)
    ensemble = {'system': prepare_system(settings),
                'engine': engine,
                'order_function': order_function,
                'rgen': create_random_generator(settings['simulation'])}
    engine.can_use_order_function(order_function)
    for key in ('steps', 'interfaces'):
        key_check(key, settings)
    controls = {'steps': settings['simulation']['steps'],
                'startcycle': settings['simulation'].get('startcycle', 0)}
    return SimulationMDFlux(ensemble, settings, controls)


def create_umbrellaw_simulation(settings):
    """Set up and create a Umbrella Window simulation.

    Parameters
    ----------
    settings : dict
        The settings needed to set up the simulation.
        Note that mincycle is trasmitted as steps to the object,
        but is has a different meaning than the other
        simulations.

    Returns
    -------
    UmbrellaWindowSimulation : obj like :py:class:`.UmbrellaWindowSimulation`
        The object representing the simulation to run.

    """
    ensemble = {'system': prepare_system(settings),
                'rgen': create_random_generator(settings['simulation'])}

    for key in ('umbrella', 'overlap', 'maxdx', 'mincycle'):
        key_check(key, settings)

    controls = {'steps': settings['simulation']['mincycle'],
                'startcycle': settings['simulation'].get('startcycle', 0)}

    return UmbrellaWindowSimulation(ensemble, settings, controls)


def create_tis_simulation(settings):
    """Set up and create a single TIS simulation.

    Parameters
    ----------
    settings : dict
        The settings needed to set up the simulation.

    Returns
    -------
    SimulationTIS : object like :py:class:`.SimulationSingleTIS`
        The object representing the simulation to run.

    """
    check_ensemble(settings)
    ensembles = create_ensembles(settings)

    key_check('steps', settings)

    controls = {'rgen': create_random_generator(settings['simulation']),
                'steps': settings['simulation']['steps'],
                'startcycle': settings['simulation'].get('startcycle', 0)}

    return SimulationTIS(ensembles, settings, controls)


def create_retis_simulation(settings):
    """Set up and create a RETIS simulation.

    Parameters
    ----------
    settings : dict
        This dictionary contains the settings for the simulation.

    Returns
    -------
    SimulationRETIS : object like :py:class:`.SimulationRETIS`
        The object representing the simulation to run.

    """
    check_ensemble(settings)
    ensembles = create_ensembles(settings)

    key_check('steps', settings)

    controls = {'rgen': create_random_generator(settings['simulation']),
                'steps': settings['simulation']['steps'],
                'startcycle': settings['simulation'].get('startcycle', 0)}

    return SimulationRETIS(ensembles, settings, controls)


def prepare_system(settings):
    """Create a system from given settings.

    Parameters
    ----------
    settings : dict
        This dictionary contains the settings for the simulation.

    Returns
    -------
    syst : object like :py:class:`.syst`
        This object will correspond to the selected simulation type.

    """
    if settings.get('system', {}).get('obj', False):
        return settings['system']['obj']

    logtxt = 'Initializing unit system.'
    print_to_screen(logtxt, level='info')
    logger.info(logtxt)
    units_from_settings(settings)

    logtxt = 'Creating system from settings.'
    print_to_screen(logtxt, level='info')
    logger.info(logtxt)
    system = create_system(settings)

    logtxt = 'Creating force field.'
    print_to_screen(logtxt, level='info')
    logger.info(logtxt)
    system.forcefield = create_force_field(settings)
    system.particles.vpot = system.evaluate_potential()

    system.extra_setup()
    return system


def prepare_engine(settings):
    """Create an engine from given settings.

    Parameters
    ----------
    settings : dict
        This dictionary contains the settings for the simulation.

    Returns
    -------
    engine : object like :py:class:`.engine`
        This object will correspond to the selected simulation type.

    """
    if settings.get('engine', {}).get('obj', False):
        return settings['engine']['obj']

    logtxt = units_from_settings(settings)
    print_to_screen(logtxt, level='info')
    logger.info(logtxt)

    check_engine(settings)
    engine = create_engine(settings)
    logtxt = f'Created engine "{engine}" from settings.'
    print_to_screen(logtxt, level='info')
    logger.info(logtxt)
    return engine


def key_check(key, settings):
    """Check for the presence of a key in settings."""
    # todo These checks shall be done earlier, when cleaning the input.
    if key not in settings['simulation']:
        msgtxt = f'Simulation setting "{key}" is missing!'
        logger.critical(msgtxt)
        raise ValueError(msgtxt)


def create_simulation(settings):
    """Create simulation(s) from given settings.

    This function will set up some common simulation types.
    It is meant as a helper function to automate some very common set-up
    task. It will here check what kind of simulation we are to perform
    and then call the appropriate function for setting that type of
    simulation up.

    Parameters
    ----------
    settings : dict
        This dictionary contains the settings for the simulation.

    Returns
    -------
    simulation : object like :py:class:`.Simulation`
        This object will correspond to the selected simulation type.

    """
    sim_type = settings['simulation']['task'].lower()

    sim_map = {
        'md': create_md_simulation,
        'md-nve': create_nve_simulation,
        'md-flux': create_mdflux_simulation,
        'umbrellawindow': create_umbrellaw_simulation,
        'make-tis-files': create_tis_simulation,
        'tis': create_tis_simulation,
        'explore': create_tis_simulation,
        'retis': create_retis_simulation,
        'pptis': create_tis_simulation,
        'repptis': create_retis_simulation,
    }

    # Improve setting quality
    add_default_settings(settings)
    add_specific_default_settings(settings)

    if settings['simulation'].get('restart', False):
        settings, info_restart = settings_from_restart(settings)

    if sim_type not in sim_map:  # TODO put in check_sim_type
        msgtxt = f'Unknown simulation task {sim_type}'
        logger.error(msgtxt)
        raise ValueError(msgtxt)

    simulation = sim_map[sim_type](settings)
    msgtxt = f'{simulation}'
    logger.info('Created simulation:\n%s', msgtxt)

    if settings['simulation'].get('restart', False):
        simulation.load_restart_info(info_restart)

    return simulation



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the Simulation class."""
from io import StringIO
import os
import logging
import unittest
import tempfile
import pathlib
import numpy as np
from pyretis.core.particles import Particles
from pyretis.core.random_gen import RandomGenerator
from pyretis.core.units import units_from_settings
from pyretis.engines.internal import Langevin
from pyretis.inout.screen import ScreenOutput
from pyretis.inout.settings import add_default_settings
from pyretis.inout.simulationio import OutputTask
from pyretis.setup.createsystem import create_system
from pyretis.simulation.simulation import Simulation
from unittest.mock import patch
from .help import turn_on_logging, TEST_SETTINGS, TxtWriter
logging.disable(logging.CRITICAL)


class TestSimulation(unittest.TestCase):
    """Run the tests for Simulation class."""

    def test_add_task(self):
        """Test that we can add task(s) to the simulation."""
        simulation = Simulation({}, {})

        def task1():  # pylint: disable=missing-docstring
            return 'Hello there!'

        def task0():  # pylint: disable=missing-docstring
            return 'I should be first!'

        task = {'func': task1, 'result': 'hello'}
        add = simulation.add_task(task)
        self.assertTrue(add)

        task = {'func': task0, 'result': 'first'}
        add = simulation.add_task(task, position=0)
        self.assertTrue(add)

        task = {'func': 100, 'result': 'should-not-be-added'}
        add = simulation.add_task(task, position=0)
        self.assertFalse(add)

    def test_run_extend(self):
        """Test that we can run and extend a simulation."""
        simulation = Simulation({}, {'steps': 10})

        def task1(simulation):
            """Dummy task for the simulation."""
            return f"Hello there {simulation.cycle['step']:03d}"

        task = {'func': task1, 'result': 'hello', 'args': (simulation,)}
        add = simulation.add_task(task)
        self.assertTrue(add)

        writer = ScreenOutput(
            TxtWriter(
                'test',
                header={'labels': ['Step', 'Message'], 'width': [9, 20]}
            )
        )
        simulation.output_tasks = [
            OutputTask('Hello-test', ['hello'], writer, {'every': 1}),
        ]

        lines = []
        with patch('sys.stdout', new=StringIO()) as stdout:
            for i, step in enumerate(simulation.run()):
                if i == 0:
                    self.assertFalse('hello' in step)
                else:
                    self.assertTrue('hello' in step)
            simulation.extend_cycles(5)
            for i, step in enumerate(simulation.run()):
                self.assertEqual(i + 1 + 10, step['cycle']['step'])
            for linei in stdout.getvalue().strip().split('\n'):
                if linei:
                    lines.append(linei)
        for i, linei in enumerate(lines):
            if i == 0:
                self.assertEqual('#    Step              Message', linei)
            else:
                msg = f'Hello there {i:03d}'
                correct = f'{i:>10d} {msg:>20s}'
                self.assertEqual(correct, linei)

    def test_restart_simple(self):
        """Test restart methods."""
        simulation = Simulation(settings={}, controls={'steps': 10})
        for _ in simulation.run():
            pass
        restart = simulation.restart_info()
        simulation2 = Simulation(settings={}, controls={'steps': 0})
        self.assertEqual(simulation2.cycle['steps'], 0)
        simulation2.load_restart_info(restart)
        self.assertEqual(simulation2.cycle['stepno'], 10)
        sr1 = simulation.restart_info()
        sr2 = simulation2.restart_info()
        # Steps and end are not copied
        self.assertNotEqual(sr1['simulation']['cycle'],
                            sr2['simulation']['cycle'])
        self.assertEqual(sr1['simulation']['type'],
                         sr2['simulation']['type'])
        # Add random generator and try again:
        rgen = RandomGenerator(seed=101)
        simulation2.load_restart_info(restart)
        # Try adding an engine when we don't expect it:
        restart['engine'] = {'rgen': rgen.get_state()}
        simulation2.load_restart_info(restart)

        # Try setting fail system:
        simulation2.system = 'bumpy'
        with self.assertRaises(AttributeError):
            simulation2.restart_info()

        # Add and try again:
        simulation2 = Simulation(settings={}, controls={'steps': 0})
        simulation2.engine = Langevin(1.0, 2.0, rgen=None)
        simulation2.load_restart_info(restart)

    def test_other_restart_features(self):
        """Test restart methods."""
        settings = TEST_SETTINGS.copy()
        add_default_settings(settings)
        simulation = Simulation(settings, controls={'steps': 0})
        simulation2 = Simulation(settings, controls={'steps': 0})
        units_from_settings(settings)
        system = create_system(settings)
        simulation.system = system
        simulation2.system = system
        particle = Particles(dim=16)
        particle.vpot = 101
        simulation.system.particles = particle

        rs1 = simulation.restart_info()
        simulation2.load_restart_info(rs1)

        self.assertTrue(np.allclose(simulation.system.particles.vpot,
                                    simulation2.system.particles.vpot))

    def test_set_up_output(self):
        """Test the set_up_output things."""
        settings = {}
        add_default_settings(settings)
        simulation = Simulation(settings, {'steps': 100})
        with tempfile.TemporaryDirectory() as tempdir:
            settings['simulation']['exe_path'] = tempdir
            simulation.set_up_output(settings)
            self.assertEqual(simulation.restart_freq,
                             settings['output']['restart-file'])
            settings['output']['restart-file'] = -1
            with turn_on_logging():
                with self.assertLogs('pyretis.simulation.simulation',
                                     level='WARNING'):
                    simulation.set_up_output(settings)
                    self.assertIsNone(simulation.restart_freq)

    def test_soft_exit(self):
        """Test the soft exit method."""
        settings = TEST_SETTINGS.copy()
        with tempfile.TemporaryDirectory() as tempdir:
            simulation = Simulation({}, {'steps': 100})
            add_default_settings(settings)
            units_from_settings(settings)
            system = create_system(settings)
            simulation.system = system
            settings['simulation']['exe_path'] = tempdir
            simulation.set_up_output(settings)
            self.assertEqual(tempdir, simulation.exe_dir)
            for i, _ in enumerate(simulation.run()):
                if i > simulation.restart_freq:
                    break
            files = [i.name for i in os.scandir(tempdir)]
            self.assertEqual(1, len(files))
            self.assertIn('pyretis.restart', files)
            exit_file = os.path.join(tempdir, 'EXIT')
            pathlib.Path(exit_file).touch()
            with patch('sys.stdout', new=StringIO()) as stdout:
                for _ in enumerate(simulation.run()):
                    pass
                self.assertIn('soft exit', stdout.getvalue().strip())
            self.assertEqual(simulation.cycle['step'],
                             simulation.restart_freq + 2)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Module for defining the generic formatting of output data from PyRETIS.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OutputFormatter (:py:class:`.OutputFormatter`)
    A generic class for formatting output from PyRETIS.

PyretisLogFormatter (:py:class:`.PyretisLogFormatter`)
    A class representing a formatter for the PyRETIS log file.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

apply_format (:py:func:`.apply_format`)
    Apply a format string to a given float value. This method
    can be used for formatting text for tables (i.e. if we want
    a fixed width).

format_number (:py:func:`.format_number`)
    Format a number based on its size.

get_log_formatter (:py:func:`.get_log_formatter`)
    Select a formatter for logging based on a given message level.

"""
import logging
from pyretis.inout.fileio import read_some_lines
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


# hard-coded formats to use for Log files:
LOG_FMT = '[%(levelname)s]: %(message)s'
LOG_DEBUG_FMT = ('[%(levelname)s] [%(name)s, %(funcName)s() at'
                 ' line %(lineno)d]: %(message)s')


__all__ = ['OutputFormatter', 'PyretisLogFormatter', 'apply_format',
           'format_number', 'get_log_formatter',
           'LOG_FMT', 'LOG_DEBUG_FMT']


def _make_header(labels, width, spacing=1):
    """Format a table header with the given labels.

    Parameters
    ----------
    labels : list of strings
        The strings to use for the table header.
    width : list of ints
        The widths to use for the table.
    spacing : int
        The spacing between the columns in the table

    Returns
    -------
    out : string
        A header for the table.

    """
    heading = []
    for i, col in enumerate(labels):
        try:
            wid = width[i]
        except IndexError:
            wid = width[-1]
        if i == 0:
            fmt = f'# {{:>{wid-2}s}}'
        else:
            fmt = f'{{:>{wid}s}}'
        heading.append(fmt.format(col))
    str_white = ' ' * spacing
    return str_white.join(heading)


def apply_format(value, fmt):
    """Apply a format string to a given float value.

    Here we check the formatting of a float. We are *forcing* a
    *maximum length* on the resulting string. This is to avoid problems
    like: '{:7.2f}'.format(12345.7) which returns '12345.70' with a
    length 8 > 7. The intended use of this function is to avoid such
    problems when we are formatting numbers for tables. Here it is done
    by switching to an exponential notation. But note however that this
    will have implications for how many decimal places we can show.

    Parameters
    ----------
    value : float
        The float to format.
    fmt : string
        The format to use.

    Note
    ----
    This function converts numbers to have a fixed length. In some
    cases this may reduce the number of significant digits. Remember
    to also output your numbers without this format in case a specific
    number of significant digits is important!

    """
    maxlen = fmt.split(':')[1].split('.')[0]
    align = ''
    if not maxlen[0].isalnum():
        align = maxlen[0]
        maxlen = maxlen[1:]
    maxlen = int(maxlen)
    str_fmt = fmt.format(value)
    if len(str_fmt) > maxlen:  # switch to exponential:
        if value < 0:
            deci = maxlen - 7
        else:
            deci = maxlen - 6
        new_fmt = f'{{:{align}{maxlen}.{deci}e}}'
        return new_fmt.format(value)
    return str_fmt


def format_number(number, minf, maxf, fmtf='{0:<16.9f}', fmte='{0:<16.9e}'):
    """Format a number based on its size.

    Parameters
    ----------
    number : float
        The number to format.
    minf : float
        If the number is smaller than `minf` then apply the
        format with scientific notation.
    maxf : float
        If the number is greater than `maxf` then apply the
        format with scientific notation.
    fmtf : string, optional
        Format to use for floats.
    fmte : string, optional
        Format to use for scientific notation.

    Returns
    -------
    out : string
        The formatted number.

    """
    if minf <= number <= maxf:
        return fmtf.format(number)
    return fmte.format(number)


class OutputFormatter:
    """A generic class for formatting output from PyRETIS.

    Attributes
    ----------
    name : string
        A string which identifies the formatter.
    header : string
        A header (or table heading) with information about the
        output data.
    print_header : boolean
        Determines if we are to print the header or not. This is useful
        for classes making use of the formatter to determine if they
        should write out the header or not.

    """

    _FMT = '{}'

    def __init__(self, name, header=None):
        """Initialise the formatter.

        Parameters
        ----------
        name : string
            A string which identifies the output type of this formatter.
        header : dict, optional
            The header for the output data

        """
        self.name = name
        self._header = None
        self.print_header = True
        if header is not None:
            if 'width' in header and 'labels' in header:
                self._header = _make_header(header['labels'],
                                            header['width'],
                                            spacing=header.get('spacing', 1))
            else:
                self._header = header.get('text', None)
        else:
            self.print_header = False

    @property
    def header(self):
        """Define the header as a property."""
        return self._header

    @header.setter
    def header(self, value):
        """Set the header."""
        self._header = value

    def format(self, step, data):
        """Use the formatter to generate output.

        Parameters
        ----------
        step : integer
            This is assumed to be the current step number for
            generating the output.
        data : list, dict or similar
            This is the data we are to format. Here we assume that
            this is something we can iterate over.

        """
        out = [f'{step}']
        for i in data:
            out.append(self._FMT.format(i))
        yield ' '.join(out)

    @staticmethod
    def parse(line):
        """Parse formatted data.

        This method is intended to be the "inverse" of the :py:meth:`.format`
        method. In this particular case, we assume that we read floats from
        columns in a file. One input line corresponds to a "row" of data.

        Parameters
        ----------
        line : string
            The string we will parse.

        Returns
        -------
        out : list of floats
            The parsed input data.

        """
        return [int(col) if i == 0 else
                float(col) for i, col in enumerate(line.split())]

    def load(self, filename):
        """Read generic data from a file.

        Since this class defines how the data is formatted it is also
        convenient to have methods for reading the data defined here.
        This method will read entire blocks of data from a file into
        memory. This will be slow for large files and this method
        could be converted to also yield the individual "rows" of
        the blocks, rather than the full blocks themselves.

        Parameters
        ----------
        filename : string
            The path/file name of the file we want to open.

        Yields
        ------
        data : list of tuples of int
            This is the data contained in the file. The columns are the
            step number, interface number and direction.

        See Also
        --------
        :py:func:`.read_some_lines`.

        """
        for blocks in read_some_lines(filename, self.parse):
            data_dict = {'comment': blocks['comment'],
                         'data': blocks['data']}
            yield data_dict

    def __str__(self):
        """Return basic info about the formatter."""
        return self.name


def get_log_formatter(level):
    """Select a log format based on a given level.

    Here, it is just used to get a slightly more verbose format for
    the debug level.

    Parameters
    ----------
    level : integer
        This integer defines the log level.

    Returns
    -------
    out : object like :py:class:`logging.Formatter`
        An object that can be used as a formatter for a logger.

    """
    if level <= logging.DEBUG:
        return PyretisLogFormatter(LOG_DEBUG_FMT)
    return PyretisLogFormatter(LOG_FMT)


class PyretisLogFormatter(logging.Formatter):  # pragma: no cover
    """Hard-coded formatter for the PyRETIS log file.

    This formatter will just adjust multi-line messages to have some
    indentation.
    """

    def format(self, record):
        """Apply the PyRETIS log format."""
        out = logging.Formatter.format(self, record)
        if '\n' in out:
            heading, _ = out.split(record.message)
            if len(heading) < 12:
                out = out.replace('\n', '\n' + ' ' * len(heading))
            else:
                out = out.replace('\n', '\n' + ' ' * 4)
        return out



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test module for the energy analysis."""
import logging
import unittest
import os
import pickle
import numpy as np
from pyretis.analysis.energy_analysis import analyse_energies
from pyretis.inout.settings import SECTIONS
from pyretis.inout.formats.energy import EnergyFile

logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


class EnergyTest(unittest.TestCase):
    """Test that we can analyse energies."""

    def test_energy_analysis(self):
        """Test the energy analysis."""
        filename = os.path.join(HERE, 'energy.txt')
        data = None
        with EnergyFile(filename, 'r') as efile:
            data = efile.load()
        settings = {
            'particles': {'npart': 1},
            'system': {'dimensions': 1, 'beta': 1, 'temperature': 1},
        }
        settings['analysis'] = SECTIONS['analysis']

        correct_file = os.path.join(HERE, 'energy-results.dat')
        for i in data:
            results = analyse_energies(i['data'], settings)
            break
        correct_data = {}
        with open(correct_file, 'rb') as infile:
            correct_data = pickle.load(infile)
        for key0 in ('time', 'temp', 'etot', 'vpot', 'ekin'):
            self.assertTrue(key0 in correct_data)
            self.assertTrue(key0 in results)
            for key, val in correct_data[key0].items():
                for i, j in zip(val, results[key0][key]):
                    self.assertTrue(np.allclose(i, j))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of creating a lattice and making a box."""
from pyretis.core import create_box
from pyretis.tools import generate_lattice
from pyretis.inout.formats.xyz import write_xyz_file

xyz, size = generate_lattice('diamond', [2, 2, 2], lcon=3.567)
write_xyz_file('diamond_small.xyz', xyz, names=['C']*len(xyz))

xyz, size = generate_lattice('diamond', [10, 10, 10], lcon=3.567)
write_xyz_file('diamond.xyz', xyz, names=['C']*len(xyz))

low = [i[0] for i in size]
high = [i[1] for i in size]
box = create_box(low=low, high=high)
print(box)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Define some common methods for the tests."""
from contextlib import contextmanager
import logging
import math
import os
import numpy as np
from numpy.random import randint, random
from pyretis.core.random_gen import MockRandomGenerator
from pyretis.core.system import System
from pyretis.core.particles import Particles, ParticlesExt
from pyretis.core.path import Path


DATA_CROSS = [[(773, 5, '+')], [(1326, 9, '-')], [(1889, 2, '+')],
              [(2761, 3, '+')], [(2966, 9, '+')], [(3233, 2, '+')],
              [(3708, 7, '-')], [(4580, 4, '+')], [(5045, 8, '+')],
              [(5672, 5, '+')]]

DATA_CROSS2 = [(773, 6, 1), (1326, 10, -1), (1889, 3, 1), (2761, 4, 1),
               (2966, 10, 1), (3233, 3, 1), (3708, 8, -1), (4580, 5, 1),
               (5045, 9, 1), (5672, 6, 1)]

CORRECT_CROSS = [
    '       773    6   +',
    '      1326   10   -',
    '      1889    3   +',
    '      2761    4   +',
    '      2966   10   +',
    '      3233    3   +',
    '      3708    8   -',
    '      4580    5   +',
    '      5045    9   +',
    '      5672    6   +',
]

DATA_ENERGY = [
    (0.873, 0.477, 0.262, 0.615),
    (0.219, 0.740, 0.786, 0.147),
    (0.235, 0.764, 0.454, 0.173),
    (0.679, 0.098, 0.130, 0.110),
    (0.442, 0.374, 0.606, 0.954),
    (0.634, 0.774, 0.168, 0.314),
    (0.732, 0.250, 0.619, 0.691),
    (0.920, 0.607, 0.515, 0.487),
    (0.856, 0.105, 0.694, 0.233),
    (0.384, 0.432, 0.528, 0.155),
]

CORRECT_ENERGY = [
    '         0       0.873000       0.477000       0.262000       0.615000',
    '         1       0.219000       0.740000       0.786000       0.147000',
    '         2       0.235000       0.764000       0.454000       0.173000',
    '         3       0.679000       0.098000       0.130000       0.110000',
    '         4       0.442000       0.374000       0.606000       0.954000',
    '         5       0.634000       0.774000       0.168000       0.314000',
    '         6       0.732000       0.250000       0.619000       0.691000',
    '         7       0.920000       0.607000       0.515000       0.487000',
    '         8       0.856000       0.105000       0.694000       0.233000',
    '         9       0.384000       0.432000       0.528000       0.155000',
]

DATA_ORDER = [
    (0.90415742, 0.99887650, 0.15510997, 0.50120026),
    (0.90798451, 0.74665316, 0.67743021, 0.88284717),
    (0.72793238, 0.26006559, 0.57081755, 0.80117034),
    (0.37518802, 0.58331773, 0.80752662, 0.19093917),
    (0.93618894, 0.01188808, 0.07401551, 0.92200559),
    (0.35298599, 0.51682127, 0.00243432, 0.63432985),
    (0.01989170, 0.86621268, 0.68238818, 0.34800920),
    (0.09492324, 0.52609810, 0.88487148, 0.56365333),
    (0.22137013, 0.63323726, 0.44261574, 0.86234165),
    (0.52925374, 0.08251509, 0.59516637, 0.87529531),
]

CORRECT_ORDER = [
    '         0     0.904157     0.998876     0.155110     0.501200',
    '         1     0.907985     0.746653     0.677430     0.882847',
    '         2     0.727932     0.260066     0.570818     0.801170',
    '         3     0.375188     0.583318     0.807527     0.190939',
    '         4     0.936189     0.011888     0.074016     0.922006',
    '         5     0.352986     0.516821     0.002434     0.634330',
    '         6     0.019892     0.866213     0.682388     0.348009',
    '         7     0.094923     0.526098     0.884871     0.563653',
    '         8     0.221370     0.633237     0.442616     0.862342',
    '         9     0.529254     0.082515     0.595166     0.875295',
]

CORRECT_PATH_EXT = [
    '# Cycle: 0, status: ACC',
    '#     Step              Filename       index    vel',
    '         0           initial.g96           0      1',
    '         1             trajB.trr           5     -1',
    '         2             trajB.trr           4     -1',
    '         3             trajB.trr           3     -1',
    '         4             trajB.trr           2     -1',
    '         5             trajB.trr           1     -1',
    '         6             trajF.trr           0      1',
    '         7             trajF.trr           1      1',
    '         8             trajF.trr           2      1',
    '         9             trajF.trr           3      1',
    '        10             trajF.trr           4      1',
]

PATH_EXT_RAW = [
    ['0', 'initial.g96', '0', '1'],
    ['1', 'trajB.trr', '5', '-1'],
    ['2', 'trajB.trr', '4', '-1'],
    ['3', 'trajB.trr', '3', '-1'],
    ['4', 'trajB.trr', '2', '-1'],
    ['5', 'trajB.trr', '1', '-1'],
    ['6', 'trajF.trr', '0', '1'],
    ['7', 'trajF.trr', '1', '1'],
    ['8', 'trajF.trr', '2', '1'],
    ['9', 'trajF.trr', '3', '1'],
    ['10', 'trajF.trr', '4', '1'],
]


@contextmanager
def turn_on_logging():
    """Turn on logging so that tests can detect it."""
    logging.disable(logging.NOTSET)
    try:
        yield
    finally:
        logging.disable(logging.CRITICAL)


def compare_lists(list1, list2, cmps='equal'):
    """Just comare some items from a list."""
    if len(list1) != len(list2):
        raise AssertionError('Different length of lists')
    if cmps == 'float':
        for i, j in zip(list1, list2):
            if not math.isclose(i, j, abs_tol=1e-8):
                raise AssertionError(f'Different {i} != {j}')
    else:
        for i, j in zip(list1, list2):
            if not i == j:
                raise AssertionError(f'Different {i} != {j}')


def assert_equal_path_dict(dict1, dict2):
    """Compare two path dictionaries."""
    comp = {
        'cycle': 'int',
        'generated': 'tuple',
        'interface': 'tuple',
        'length': 'int',
        'ordermax': 'tuple-float',
        'ordermin': 'tuple-float',
        'status': 'string',
        'weight': 'float'
    }
    for key in dict1:
        if key not in dict2:
            raise AssertionError('Different keys in dicts!')
    for key in dict2:
        if key not in dict1:
            raise AssertionError('Different keys in dics!')
    for key, val in comp.items():
        if key not in dict1 or key not in dict2:
            raise AssertionError(f'Missing key: {key}')
        if val in ('int', 'string', 'float'):
            if not dict1[key] == dict2[key]:
                raise AssertionError(f'Different {key}')
        elif val in ('tuple', 'tuple-float'):
            if val != 'tuple-float':
                compare_lists(dict1[key], dict2[key])
            else:
                compare_lists(dict1[key], dict2[key], cmps='float')


def _add_path_data(path_data, phasepoint):
    """Add path data for storage."""
    if not path_data:
        step = 0
    else:
        step = path_data[-1][0] + 1
    new_data = [step, phasepoint.particles.get_pos()[0],
                phasepoint.particles.get_pos()[1],
                phasepoint.particles.get_vel()]
    path_data.append(new_data)


def set_up_system(order, pos, vel, vpot=None, ekin=None, internal=True):
    """Create a system for testing."""
    system = System()
    if internal:
        system.particles = Particles(dim=3)
        for posi, veli in zip(pos, vel):
            system.add_particle(posi, vel=veli)
    else:
        system.particles = ParticlesExt(dim=3)
        system.add_particle(pos, vel=vel)
    system.order = order
    system.particles.vpot = vpot
    system.particles.ekin = ekin
    return system


def create_external_path(random_length=False):
    """Create an external path for testing."""
    path_data = []
    path = Path(None)
    phasepoint = set_up_system([random(), random()],
                               ('initial.g96', None),
                               False, internal=False)
    path.append(phasepoint)
    _add_path_data(path_data, phasepoint)
    if random_length:
        length = randint(10, 100)
    else:
        length = 5
    for i in range(length, 0, -1):
        phasepoint = set_up_system([random(), random()],
                                   ('trajB.trr', i), True,
                                   vpot=random(), ekin=random(),
                                   internal=False)
        path.append(phasepoint)
        _add_path_data(path_data, phasepoint)
    for i in range(0, length):
        phasepoint = set_up_system([random(), random()],
                                   ('trajF.trr', i), False,
                                   vpot=random(), ekin=random(),
                                   internal=False)
        path.append(phasepoint)
        _add_path_data(path_data, phasepoint)
    return path, path_data


def create_test_paths(npath=5):
    """Create some paths we can use.

    Parameters
    ----------
    npath : integer
        The number of paths to create.

    """
    paths = []
    correct_data = []
    for i in range(npath):
        length = randint(10, 100)
        path = Path(rgen=MockRandomGenerator(seed=0))
        data_copy = {'order': [], 'pos': [], 'vel': [],
                     'ekin': [], 'vpot': []}
        for j in range(length):
            phasepoint = set_up_system([(i + 1) * j, j, j*10],
                                       np.ones((9, 3)) * i * j,
                                       np.ones((9, 3)) * i * j * -1,
                                       vpot=(i + j),
                                       ekin=(i + j + 1))
            path.append(phasepoint)
            data_copy['order'].append(phasepoint.order)
            for key in ('pos', 'vel', 'ekin', 'vpot'):
                data_copy[key].append(getattr(phasepoint.particles, key))
        paths.append(path)
        correct_data.append(data_copy)
    return paths, correct_data


def remove_file(filename):
    """Remove the given file."""
    try:
        os.remove(filename)
    except OSError:
        pass



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""The Ackley function, used as a potential."""
import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=unused-import
from pyretis.forcefield import PotentialFunction


TWO_PI = np.pi * 2.0
EXP = np.exp(1)


@np.vectorize
def ackley_potential(x, y):  # pylint: disable=invalid-name
    """Evaluate the Ackley function."""
    return (-20.0 * np.exp(-0.2*np.sqrt(0.5*(x**2 + y**2))) -
            np.exp(0.5 * (np.cos(TWO_PI * x) + np.cos(TWO_PI * y))) +
            EXP + 20)


class Ackley(PotentialFunction):
    """A implementation of the Ackley function.

    Note that the usage of this potential function differs from
    the usual usage for force fields.

    """

    def __init__(self):
        """Set up the function."""
        super().__init__(dim=2, desc='The Ackley function')

    def potential(self, system):
        """Evaluate the potential, note that we return all values."""
        xpos = system.particles.pos[:, 0]
        ypos = system.particles.pos[:, 1]
        pot = ackley_potential(xpos, ypos)
        return pot


def main():
    """Plot the Ackley function."""
    xgrid, ygrid = np.meshgrid(np.linspace(-5, 5, 100),
                               np.linspace(-5, 5, 100))
    zgrid = ackley_potential(xgrid, ygrid)

    fig = plt.figure()
    ax1 = fig.add_subplot(211)
    ax2 = fig.add_subplot(212, projection='3d')
    ax1.contourf(xgrid, ygrid, zgrid)
    ax2.plot_surface(xgrid, ygrid, zgrid, cmap=plt.get_cmap('viridis'))
    plt.show()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains functions that act on (a selection of) particles.

These functions are intended for calculating particle properties as the
kinetic temperature, pressure etc.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

atomic_kinetic_energy_tensor (:py:func:`.atomic_kinetic_energy_tensor`)
    Return the kinetic energy tensor for each atom in a selection
    of particles.

calculate_kinetic_energy (:py:func:`.calculate_kinetic_energy`)
    Return the kinetic energy of a collection of particles.

calculate_kinetic_energy_tensor (:py:func:`.calculate_kinetic_energy_tensor`)
    Return the kinetic energy tensor for a selection of particles.

calculate_kinetic_temperature (:py:func:`calculate_kinetic_temperature`)
    Return the kinetic temperature of a collection of particles.

calculate_linear_momentum (:py:func:`calculate_linear_momentum`)
    Calculates the linear momentum of a collection of particles.

calculate_pressure_from_temp (:py:func:`calculate_pressure_from_temp`)
    Return the scalar pressure using the temperature and the virial.

calculate_pressure_tensor (:py:func:`calculate_pressure_tensor`)
    Return the pressure tensor, obtained from the virial and the kinetic
    energy tensor.

calculate_scalar_pressure (:py:func:`calculate_scalar_pressure`)
    Return the scalar pressure (from the trace of the pressure tensor).

calculate_thermo (:py:func:`calculate_thermo`)
    Calculate and return several "thermodynamic" properties as the
    potential, kinetic and total energies per particle, the temperature,
    the pressure and the momentum.

calculate_thermo_path (:py:func:`calculate_thermo_path`)
    Calculate and return some thermodynamic properties. This method
    is similar to the `calculate_thermo`, however, it is simpler and
    calculates fewer quantities.

kinetic_energy (:py:func:`kinetic_energy`)
    Return the kinetic energy for velocities and masses given as
    numpy arrays.

kinetic_temperature (:py:func:`kinetic_temperature`)
    Return the temperature for velocities and masses given as
    numpy arrays.

reset_momentum (:py:func:`reset_momentum`)
    Set linear momentum (for a selection of particles) to zero.
"""
import numpy as np

np.set_printoptions(legacy='1.21')

__all__ = [
    'atomic_kinetic_energy_tensor',
    'calculate_kinetic_energy',
    'calculate_kinetic_energy_tensor',
    'calculate_kinetic_temperature',
    'calculate_linear_momentum',
    'calculate_pressure_from_temp',
    'calculate_pressure_tensor',
    'calculate_scalar_pressure',
    'calculate_thermo',
    'calculate_thermo_path',
    'kinetic_energy',
    'kinetic_temperature',
    'reset_momentum'
]


def _get_vel_mass(particles, selection=None):
    """Return velocity and mass for a selection.

    This is just for convenience since we are using this
    selection a lot.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    selection : list of integers, optional
        A list with indices of particles to use in the calculation.

    Returns
    -------
    out[0] : numpy.array
        The velocities corresponding to the selection.
    out[1] : numpy.array, optional
        The masses corresponding to the selection.

    """
    if selection is None:
        vel = particles.vel
        mass = particles.mass
    else:
        vel = particles.vel[selection]
        mass = particles.mass[selection]
    return vel, mass


def atomic_kinetic_energy_tensor(particles, selection=None):
    """Return kinetic energy tensors for a particle selection.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    selection : list of integers, optional
        A list with indices of particles to use in the calculation.

    Returns
    -------
    kin : numpy.array
        A numpy array with dimensionality equal to
        (`len(selection)`, `dim`, `dim`) where `dim` is the number of
        dimensions used in the velocities. `kin[i]` contains the kinetic
        energy tensor formed by the outer product of `mol[selection][i]`
        and `vel[selection][i]`. The sum of the tensor should equal the
        output from `calculate_kinetic_energy_tensor`.

    """
    vel, mass = _get_vel_mass(particles, selection=selection)
    mom = vel * mass
    if len(mass) == 1:  # in general: selection != particles.npart
        kin = 0.5 * np.outer(mom, vel)
    else:
        kin = 0.5 * np.einsum('ij,ik->ijk', mom, vel)
    return kin


def calculate_kinetic_energy(particles, selection=None, kin_tensor=None):
    """Return the kinetic energy of a collection of particles.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    selection : list of integers, optional
        A list with indices of particles to use in the calculation.
    kin_tensor : numpy.array, optional
        If kinetic_tensor is not given, the kinetic energy tensor will
        be calculated.

    Returns
    -------
    out[0] : float
        The scalar kinetic energy.
    out[1] : numpy.array
        The kinetic energy tensor.

    """
    if kin_tensor is None:
        kin_tensor = calculate_kinetic_energy_tensor(particles,
                                                     selection=selection)
    return kin_tensor.trace(), kin_tensor


def calculate_kinetic_energy_tensor(particles, selection=None):
    """Return the kinetic energy tensor for a selection of particles.

    The tensor is formed as the outer product of the velocities.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    selection : list of integers, optional
        A list with indices of particles to use in the calculation.

    Returns
    -------
    out : numpy.array
        The kinetic energy tensor. Dimensionality is equal to (dim, dim)
        where dim is the number of dimensions used in the velocities.
        The trace gives the kinetic energy.

    """
    vel, mass = _get_vel_mass(particles, selection=selection)
    _, kin = kinetic_energy(vel, mass)
    return kin


def kinetic_energy(vel, mass):
    """Obtain the kinetic energy for given velocities and masses.

    Parameters
    ----------
    vel : numpy.array
        The velocities
    mass : numpy.array
        The masses. This is assumed to be a column vector.

    Returns
    -------
    out[0] : float
        The kinetic energy
    out[1] : numpy.array
        The kinetic energy tensor.

    """
    mom = vel * mass
    if len(mass) == 1:
        kin = 0.5 * np.outer(mom, vel)
    else:
        kin = 0.5 * np.einsum('ij,ik->jk', mom, vel)
    return kin.trace(), kin


def calculate_kinetic_temperature(particles, boltzmann, dof=None,
                                  selection=None,
                                  kin_tensor=None):
    """Return the kinetic temperature of a collection of particles.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    boltzmann : float
        This is the Boltzmann factor/constant in correct units.
    dof : list of floats, optional
        The degrees of freedom to subtract. Its shape should
        be equal to the number of dimensions.
    selection : list of integers, optional
        A list with indices of particles to use in the calculation.
    kin_tensor : numpy.array optional
        The kinetic energy tensor. If the kinetic energy tensor is not
        given, it will be recalculated here.

    Returns
    -------
    out[0] : numpy.array
        Array with the same size as the kinetic energy. It
        contains the temperature in each spatial dimension.
    out[1] : float
        The temperature averaged over all dimensions.
    out[2] : numpy.array
        The kinetic energy tensor.

    """
    vel, mass = _get_vel_mass(particles, selection=selection)
    npart = len(mass)  # using mass, since selection may be != particles.npart
    ndof = npart * np.ones(vel[0].shape)

    if kin_tensor is None:
        kin_tensor = calculate_kinetic_energy_tensor(particles,
                                                     selection=selection)
    if dof is not None:
        ndof = ndof - dof
    temperature = (2.0 * kin_tensor.diagonal() / ndof) / boltzmann
    return temperature, np.average(temperature), kin_tensor


def kinetic_temperature(vel, mass, boltzmann, dof=None, kin_tensor=None):
    """Return the kinetic temperature given velocities and masses.

    This method does not work on a particle object, but rather with
    numpy arrays. That is, it is intended for use when we can't rely
    on the particle object.

    Parameters
    ----------
    vel : numpy.array
        The velocities
    mass : numpy.array
        The masses. This is assumed to be a column vector.
    boltzmann : float
        This is the Boltzmann factor/constant in correct units.
    dof : list of floats, optional
        The degrees of freedom to subtract. Its shape should
        be equal to the number of dimensions.
    kin_tensor : numpy.array optional
        The kinetic energy tensor. If the kinetic energy tensor is not
        given, it will be recalculated here.

    Returns
    -------
    out[0] : numpy.array
        It contains the temperature in each spatial dimension.
        Array with the same size as the kinetic energy.
    out[1] : float
        The temperature averaged over all dimensions.
    out[2] : numpy.array
        The kinetic energy tensor.

    """
    npart = len(mass)  # using mass, since selection may be != particles.npart
    ndof = npart * np.ones(vel[0].shape)

    if kin_tensor is None:
        _, kin_tensor = kinetic_energy(vel, mass)

    if dof is not None:
        ndof = ndof - dof
    temperature = (2.0 * kin_tensor.diagonal() / ndof) / boltzmann
    return temperature, np.average(temperature), kin_tensor


def calculate_linear_momentum(particles, selection=None):
    """Calculate the linear momentum for a collection of particles.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    selection : list of integers, optional
        A list with indices of particles to use in the calculation.

    Returns
    -------
    out : numpy.array
        The array contains the linear momentum for each dimension.

    """
    vel, mass = _get_vel_mass(particles, selection=selection)
    return np.sum(vel * mass, axis=0)


def calculate_pressure_from_temp(particles, dim, boltzmann, volume,
                                 dof=None):
    """Evaluate the scalar pressure.

    The scalar pressure is here calculated  using the temperature
    and the degrees of freedom.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    dim : int
        This is the dimensionality of the system.
        Typically provided by `system.get_dim()`.
    boltzmann : float
        This is the Boltzmann factor/constant in correct units.
        Typically it can be supplied by `system.get_boltzmann()`.
    volume : float
        This is the volume 'occupied' by the particles. It can typically
        be obtained by a `box.calculate_volume()`
    dof : list of floats, optional
        The degrees of freedom to subtract. Its shape should
        be equal to the number of dimensions.

    Returns
    -------
    out[0] : float
        Pressure times volume.
    out[1] : float
        The pressure.

    Note
    ----
    This function may possibly be removed - it does not appear to be
    very useful right now.

    """
    if dof is None:
        ndof = particles.npart
    else:
        ndof = (particles.npart * dim - np.sum(dof)) / float(dim)
    _, temperature, _ = calculate_kinetic_temperature(particles, boltzmann,
                                                      dof=dof)
    pressvolume = ndof * temperature * boltzmann
    pressvolume += particles.virial.trace() / float(dim)
    press = pressvolume / volume
    return pressvolume, press


def calculate_pressure_tensor(particles, volume, kin_tensor=None):
    """Calculate the pressure tensor.

    The pressure tensor is obtained from the virial the kinetic
    energy tensor.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    volume : float
        This is the volume 'occupied' by the particles. It can typically
        be obtained by a `box.calculate_volume()`.
    kin_tensor : numpy.array, optional
        The kinetic energy tensor. If `kin_tensor` is not given, it will
        be calculated here.

    Returns
    -------
    out : numpy.array
        The symmetric pressure tensor, dimensions (`dim`, `dim`), where
        `dim` = the number of dimensions considered in the simulation.

    """
    if kin_tensor is None:
        kin_tensor = calculate_kinetic_energy_tensor(particles, selection=None)
    pressure = (particles.virial + 2. * kin_tensor) / volume
    return pressure


def calculate_scalar_pressure(particles, volume, dim, press_tensor=None,
                              kin_tensor=None):
    """Evaluate the scalar pressure using the pressure tensor.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    volume : float
        This is the volume 'occupied' by the particles. It can typically
        be obtained by a `box.calculate_volume()`.
    dim : int
        This is the dimensionality of the system. Typically provided by
        `system.get_dim()`
    press_tensor : numpy.array, optional
        If `press_tensor` is not given, the pressure tensor will be
        calculated here.
    kin_tensor : numpy.array, optional
        If `kin_tensor` is not given, the kinetic energy tensor will be
        calculated here.

    Returns
    -------
    out : float
        The scalar pressure averaged over the diagonal components of
        the pressure tensor.

    """
    if press_tensor is None:
        press_tensor = calculate_pressure_tensor(particles, volume,
                                                 kin_tensor=kin_tensor)
    return press_tensor.trace() / float(dim)


def calculate_thermo(system, dof=None, dim=None, volume=None, vpot=None):
    """Calculate and return several thermodynamic properties.

    The calculated properties are the potential, kinetic and total
    energies per particle, the temperature, the pressure and the
    momentum.

    Parameters
    ----------
    system : object like :py:class:`.System`
        This object is used to access the particles and the box.
    dof : list of floats, optional
        The degrees of freedom.
    dim : float, optional
        The dimensionality of, typically provided with a
        `system.get_dim()`.
    volume : float, optional
        This is the volume 'occupied' by the particles. It can typically
        be obtained by a `system.box.calculate_volume()`.
    vpot : float, optional
        The potential energy of the particles.

    Returns
    -------
    out : dict
        This dict contains the float that is calculated in this routine.

    """
    if volume is None:
        volume = system.box.calculate_volume()
    if vpot is None:
        vpot = system.particles.vpot
    if dim is None:
        dim = system.get_dim()
    if dof is None:
        dof = system.temperature['dof']
    kin_tens = calculate_kinetic_energy_tensor(system.particles)
    _, temp, _ = calculate_kinetic_temperature(system.particles,
                                               system.get_boltzmann(),
                                               dof=dof,
                                               kin_tensor=kin_tens)
    press_tens = calculate_pressure_tensor(system.particles, volume,
                                           kin_tensor=kin_tens)
    ekin = kin_tens.trace()
    press = calculate_scalar_pressure(system.particles, volume, dim,
                                      kin_tensor=kin_tens)
    mom = calculate_linear_momentum(system.particles)
    npart = float(system.particles.npart)
    result = {'vpot': vpot / npart, 'ekin': ekin / npart,
              'etot': (ekin + vpot) / npart,
              'temp': temp, 'press': press, 'mom': mom,
              'press-tens': press_tens}
    return result


def calculate_thermo_path(system):
    """Calculate and return several thermodynamic properties.

    The calculated properties are the potential, kinetic and total
    energies for the system and the current temperature. The name here
    ``calculate_thermo_path`` just indicates that this function is
    useful in connection with path sampling simulations, i.e. it
    just calculates a few energy terms.

    Parameters
    ----------
    system : object like :py:class:`.System`
        This object is used to access the particles and the box.

    Returns
    -------
    out : dict
        This dict contains the float that is calculated in this routine.

    """
    particles = system.particles
    kin_tens = calculate_kinetic_energy_tensor(particles)
    _, temp, _ = calculate_kinetic_temperature(particles,
                                               system.get_boltzmann(),
                                               dof=system.temperature['dof'],
                                               kin_tensor=kin_tens)
    ekin = kin_tens.trace()
    vpot = particles.vpot
    return {'vpot': vpot, 'ekin': ekin, 'etot': ekin + vpot, 'temp': temp}


def reset_momentum(particles, selection=None, dim=None):
    """Set the linear momentum of a selection of particles to zero.

    Parameters
    ----------
    particles : object like :py:class:`.Particles`
        This object represents the particles.
    selection : list of integers, optional
        A list with indices of particles to use in the calculation.
    dim : list or None, optional
        If ``dim`` is None, the momentum will be reset for ALL
        dimensions. Otherwise, it will only be applied to the
        dimensions where ``dim`` is True.

    Returns
    -------
    out : None
        Returns `None` and modifies velocities of the selected
        particles.

    """
    vel, mass = _get_vel_mass(particles, selection=selection)
    mom = np.sum(vel * mass, axis=0)
    if dim is not None:
        for i, reset in enumerate(dim):
            if not reset:
                mom[i] = 0
    particles.vel[selection] -= (mom / mass.sum())



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module just contains some info for PyRETIS.

Here we define the name of the program and some other
relevant info.

This file is generated by PyRETIS (``setup_version.py``)
"""
PROGRAM_NAME = 'PyRETIS'
URL = 'http://www.pyretis.org'
GIT_URL = 'https://gitlab.com/pyretis/pyretis'
CITE = """
[1] A. Lervik, E. Riccardi and T. S. van Erp, J. Comput. Chem., 2017
    doi: https://dx.doi.org/10.1002/jcc.24900
[2] E. Riccardi, A. Lervik, S. Roet, O. Aarøen and T. S. van Erp,
    J. Comput. Chem., 2019, doi: https://dx.doi.org/10.1002/jcc.26112
[3] W. Vervust, D. T. Zhang, A. Ghysels, S. Roet, T. S. van Erp, E. Riccardi,
    J. Comput. Chem. 2024, 1. https://doi.org/10.1002/jcc.27319

"""
LOGO = r"""

  _______                                     _______________
 /   ___ \       _____  ____  ____________   / __        ___/
/_/\ \_/ /_  __  \  _ \/ __/ /_  _/ / ___/  /_/ / // // /
   /  __  / / /  / /_)  _/    / // /\ \        / // // / __
  / /   \ \/ /  / _  \ /__   / // /__\ \   ___/ // // /_/ /
 /_/    _\  /  /_/ |_|___/  /_//_/_____/  /______________/
       /___/
"""



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of a class for simulation tasks.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

SimulationTask (:py:class:`.SimulationTask`)
    A class representing a simulation task.

SimulationTaskList (:py:class:`.SimulationTaskList`)
    A class for representing a list of simulation tasks. This class
    defines functionality for adding tasks from a dictionary description.

"""
import logging
from pyretis.core.common import inspect_function
from pyretis.inout.simulationio import Task
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


def _check_args(function, given_args=None, given_kwargs=None):
    """Check consistency for function and the given (keyword) arguments.

    Here we assume that the arguments are given in a list and that
    the keyword arguments are given as a dictionary. The function
    `inspect.getargspec` is used to check the input function.

    Parameters
    ----------
    function : callable
        The function we will inspect.
    given_args : list, optional
        A list of the arguments to pass to the function. 'self' will not
        be considered here since it passed implicitly.
    given_kwargs : dict, optional
        A dictionary with keyword arguments.

    Returns
    -------
    out : boolean
        False if there is some inconsistencies, i.e. when the calling
        of the given `function` will probably fail. True otherwise.

    """
    arguments = inspect_function(function)
    args = [arg for arg in arguments['args'] if arg != 'self']
    defaults = list(arguments['kwargs'])
    # first test, do we give correct number of required arguments?
    if given_args is not None:
        given = len(given_args)
    else:
        given = 0
    if len(args) != given:
        msgtxt = 'Wrong number of arguments given'
        logger.warning(msgtxt)
        return False
    # Check kwargs but only check in case some kwargs are given here.
    # If they are not given, we assume that the user knows what's happening
    # and that the default kwargs will be used.
    if given_kwargs is not None:
        if defaults:
            extra = [key for key in given_kwargs if key not in defaults]
            if extra:
                msg = [f'Task Keyword arguments: {defaults}']
                msg += [f'Unexpected keyword argument: {extra}']
                msgtxt = '\n'.join(msg)
                logger.warning(msgtxt)
                return False
        else:
            msgtxt = 'Unexpected keyword argument!'
            logger.warning(msgtxt)
            return False
    return True


class SimulationTask(Task):
    """Representation of simulation tasks.

    This class defines a task object. A task is executed at specific
    points, at regular intervals etc. in a simulation. A task will
    typically provide a result, but it does not need to. It can simply
    just alter the state of the passed argument(s).

    Attributes
    ----------
    function : function
        The function to execute.
    when : dict
        Determines when the task should be executed.
    args : list
        List of arguments to the function.
    kwargs : dict
        The keyword arguments to the function.
    first : boolean
        True if this task should be executed before the first
        step of the simulation.
    result : string
        This is a label for the result created by the task.

    """

    def __init__(self, function, args=None, kwargs=None, when=None,
                 result=None, first=False):
        """Initialise the task.

        Parameters
        ----------
        function : callable
            The function to execute.
        args : list, optional
            List of arguments to the function.
        kwargs : dict, optional
            The keyword arguments to the function.
        when : dict, optional
            Determines if the task should be executed.
        result : string, optional
            This is a label for the result created by the task.
        first : boolean, optional
            True if this task should be executed before the first
            step of the simulation.

        """
        if not callable(function):
            msg = 'The given function for the task is not callable!'
            raise AssertionError(msg)
        ok_to_add = _check_args(function, given_args=args, given_kwargs=kwargs)
        if not ok_to_add:
            msg = 'Wrong arguments or keyword arguments!'
            raise AssertionError(msg)
        super().__init__(when)
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self._result = result
        self.first = first

    def execute(self, step):
        """Execute the task.

        Parameters
        ----------
        step : dict of ints
            The keys are:

            * 'step': the current cycle number.
            * 'startcycle': the cycle number at the start.
            * 'stepno': the number of cycles we have performed so far.

        Returns
        -------
        out : unknown type
            The result of running `self.function`.

        """
        args = self.args
        kwargs = self.kwargs
        if self.execute_now(step):
            if args is None:
                if kwargs is None:
                    return self.function()
                return self.function(**kwargs)
            if kwargs is None:
                return self.function(*args)
            return self.function(*args, **kwargs)
        return None

    @property
    def result(self):
        """Return the result label."""
        return self._result

    def run_first(self):
        """Return True if task should be executed before first step."""
        return self.first

    def task_dict(self):
        """Return a dict representing the task."""
        return {'func': self.function, 'args': self.args,
                'kwargs': self.kwargs, 'when': self.when,
                'result': self.result, 'first': self.first,
                'func-name': self.function.__name__}

    def __call__(self, step):
        """Execute the task.

        Parameters
        ----------
        step : dict of ints
            The keys are:

            * 'step': the current cycle number.
            * 'startcycle': the cycle number at the start.
            * 'stepno': the number of cycles we have performed so far.

        Returns
        -------
        out : unknown type
            The result of `self.execute(step)`.

        """
        return self.execute(step)

    def __str__(self):
        """Output info about the task."""
        msg = ['Task:']
        msg += [f' -> Function name: {self.function.__name__}']
        msg += [f' -> Function args: {self.args}']
        msg += [f' -> Function kwargs: {self.kwargs}']
        msg += [f' -> Execute when: {self.when}']
        msg += [f' -> Execute at first: {self.first}']
        msg += [f' -> Result: {self._result}']
        return '\n'.join(msg)
